/*
 * CUDA for Sparse Solver
 * Author: Luo Li     li.luo@sub.siat.ac.cn
 */ 

#ifndef __JACOBI_PRECON_H__
#define __JACOBI_PRECON_H__

template <class T>
struct Jacobi_precon
{
    int nrow;
    T *diag;
};

template <class T>
double get_precon_flop(Jacobi_precon<T> *h_precon)
{
    int n = h_precon->nrow;
    return double(n);
}

template <class T>
void Jacobi_setup(spm_diag<T> *spm_dia, Jacobi_precon<T> *Jacobi)
{
    Jacobi->nrow = spm_dia->nrow;
    #if USE_PAGE_LOCKED
    CUDA_SAFE_CALL(cudaMallocHost((void **)&Jacobi->diag, Jacobi->nrow * sizeof(T)));
    #else
    Jacobi->diag = (T *)malloc(sizeof(T) * Jacobi->nrow);
    #endif
    
    for(int i = 0; i < Jacobi->nrow; i++)
    {  
        T val_temp = (spm_dia->diag)[i];
        (Jacobi->diag)[i] = (val_temp == 0.0) ? 1.0 : 1.0 / val_temp;
    }    
}

template <class T>
void clear_precon(Jacobi_precon<T> *Jacobi) 
{
    #if USE_PAGE_LOCKED
    CUDA_SAFE_CALL(cudaFreeHost(Jacobi->diag));
    #else
    free(Jacobi->diag);
    #endif
}

template <class T>
Jacobi_precon<T> d_alloc_precon(Jacobi_precon<T> *h_precon)
{
    Jacobi_precon<T> d_precon;
    d_precon.nrow = h_precon->nrow;

    CUDA_SAFE_CALL(cudaMalloc((void **)&d_precon.diag, d_precon.nrow * sizeof(T)));
    CUDA_SAFE_CALL(cudaMemcpy(d_precon.diag, h_precon->diag, d_precon.nrow * sizeof(T), cudaMemcpyHostToDevice));

    return d_precon;    
}

template <class T>
void d_dealloc_precon(Jacobi_precon<T> *d_precon)
{
    CUDA_SAFE_CALL(cudaFree(d_precon->diag));
}

template <class T>
Jacobi_precon<T> md_alloc_precon(Jacobi_precon<T> *h_precon, const int n_per_thr, const int thread_start)
{
    Jacobi_precon<T> md_precon;
    md_precon.nrow = h_precon->nrow;

    cublasAlloc(n_per_thr, sizeof(T), (void **)&md_precon.diag);
    cublasSetVector(n_per_thr, sizeof(T), h_precon->diag + thread_start, 1, md_precon.diag, 1);

    return md_precon;    
}

template <class T>
void md_dealloc_precon(Jacobi_precon<T> *md_precon)
{
    cublasFree(md_precon->diag);
}

template <class T>
void precon_Jacobi_cpu(T *diag, T *r, T *h, const int n)
{
    for(int i = 0; i < n; i++)
        h[i] = r[i] * diag[i];
}

template <class Matrix_T, class T>
void precon_cpu(Matrix_T *h_spm, Jacobi_precon<T> *h_precon, T *r, T *h)
{
    int n = h_spm->ncol;
    #if USE_FLOAT_PRECON
    float *diag_float = (float *)malloc(n * sizeof(float));
    float *r_float = (float *)malloc(n * sizeof(float));
    float *h_float = (float *)malloc(n * sizeof(float));
    transform(h_precon->diag, h_precon->diag + n, diag_float, convert<T, float>());
    transform(r, r + n, r_float, convert<T, float>());
    transform(h, h + n, h_float, convert<T, float>());
    precon_Jacobi_cpu<float>(diag_float, r_float, h_float, n);
    transform(h_float, h_float + n, h, convert<float, T>());
    free(diag_float);
    free(r_float);
    free(h_float);
    #else
    precon_Jacobi_cpu<T>(h_precon->diag, r, h, n);
    #endif    
}

template <class Matrix_T, class T>
void precon_single_gpu(Matrix_T *d_spm, Jacobi_precon<T> *d_precon, Jacobi_precon<T> *h_precon, T *r, T *h)
{
    int n = d_spm->ncol;
    unsigned int dimBlock = DIMBLOCK;
    const dim3 dimGrid = large_grid(n, dimBlock); 

    #if USE_FLOAT_PRECON
    float *diag_float = 0;
    float *r_float = 0;
    float *h_float = 0;
    CUDA_SAFE_CALL(cudaMalloc((void **)&diag_float, n * sizeof(float)));
    CUDA_SAFE_CALL(cudaMalloc((void **)&r_float, n * sizeof(float)));
    CUDA_SAFE_CALL(cudaMalloc((void **)&h_float, n * sizeof(float)));
    transform_gpu<T, float><<<dimGrid, dimBlock>>>(d_precon->diag, diag_float, n);
    transform_gpu<T, float><<<dimGrid, dimBlock>>>(r, r_float, n);
    transform_gpu<T, float><<<dimGrid, dimBlock>>>(h, h_float, n);
    #if USE_TEXTURE
    bind_vec(r_float);
    #endif 
    precon_Jacobi_gpu<float><<<dimGrid, dimBlock>>>(diag_float, r_float, h_float, n);    
    #if USE_TEXTURE
    unbind_vec(r_float);
    #endif 
    transform_gpu<float, T><<<dimGrid, dimBlock>>>(h_float, h, n);
    cublasFree(diag_float);
    cublasFree(r_float);
    cublasFree(h_float);
    #else
    #if USE_TEXTURE
    bind_vec(r);
    #endif
    precon_Jacobi_gpu<T><<<dimGrid, dimBlock>>>(d_precon->diag, r, h, n);    
    #if USE_TEXTURE
    unbind_vec(r);
    #endif     
    #endif     
}

template <class Matrix_T, class T>
void precon_multi_gpu(Matrix_T *md_spm, Jacobi_precon<T> *md_precon, T *r, T *h, const int n_per_thr)
{
    unsigned int dimBlock = DIMBLOCK;
    const dim3 dimGrid = large_grid(n_per_thr, dimBlock);    
    
    #if USE_FLOAT_PRECON
    float *diag_float = 0;
    float *r_float = 0;
    float *h_float = 0;
    CUDA_SAFE_CALL(cudaMalloc((void **)&diag_float, n_per_thr * sizeof(float)));
    CUDA_SAFE_CALL(cudaMalloc((void **)&r_float, n_per_thr * sizeof(float)));
    CUDA_SAFE_CALL(cudaMalloc((void **)&h_float, n_per_thr * sizeof(float)));
    transform_gpu<T, float><<<dimGrid, dimBlock>>>(md_precon->diag, diag_float, n_per_thr);
    transform_gpu<T, float><<<dimGrid, dimBlock>>>(r, r_float, n_per_thr);
    transform_gpu<T, float><<<dimGrid, dimBlock>>>(h, h_float, n_per_thr);
    #if USE_TEXTURE
    bind_vec(r_float);
    #endif    
    precon_Jacobi_gpu<float><<<dimGrid, dimBlock>>>(diag_float, r_float, h_float, n_per_thr);    
    #if USE_TEXTURE
    unbind_vec(r_float);
    #endif     
    transform_gpu<float, T><<<dimGrid, dimBlock>>>(h_float, h, n_per_thr);
    cublasFree(diag_float);
    cublasFree(r_float);
    cublasFree(h_float);
    #else
    #if USE_TEXTURE
    bind_vec(r);
    #endif    
    precon_Jacobi_gpu<T><<<dimGrid, dimBlock>>>(md_precon->diag, r, h, n_per_thr);               
    #if USE_TEXTURE
    unbind_vec(r);
    #endif     
    #endif
}

#endif /* __JACOBI_PRECON_H__ */
