/*
 * CUDA for Sparse Solver
 * Author: Luo Li     li.luo@sub.siat.ac.cn
 */ 

#ifndef __CG_HOST_H__
#define __CG_HOST_H__

#include "blas.h"

template <class Matrix_T, class T>
T cg_host(FILE *fpDat, Matrix_T *h_spm, T *h_b, T *h_x, const int max_iter, const T error_rel, int *iter)
{     
    int n;
    if(h_spm->nrow == h_spm->ncol) 
        n = h_spm->nrow;
    else
    {
        printf("CG_HOST: Nonsquare matrix doesn't supported.\n");
        exit(EXIT_FAILURE);
    } 

    int k;
    T alpha, beta, error_norm, r2;

    unsigned int timer_compu = 0;
    double t_compu_ms = 0.0;    
    start_timer(&timer_compu);    

    /* allocate temporary vectors */
    T *p = (T *)malloc(n * sizeof(T));
    T *r = (T *)malloc(n * sizeof(T));
    T *Ap = (T *)malloc(n * sizeof(T));

    #if USE_ERROR_REL
    T b_square = Xnrm2(n, h_b, 1); 
    T error_norm_rel = error_rel * error_rel * b_square;
    #endif
	
	/* r := b - Ax */
    spmv_cpu<T>(h_spm, h_x, r);	
    Xaxpy(n, -1.0, h_b, 1, r, 1);
    Xscal(n, -1.0, r, 1);    
	/* p := r */
    Xcopy(n, r, 1, p, 1);
	error_norm = Xnrm2(n, r, 1); 
	
    #if SET_SUPER_MAX_ITER
    for(k = 0; k < SUPER_MAX_ITER; ++k)
    #else
    for(k = 0; k < n; ++k)
    #endif
	{
		/* alpha := r'r / p'Ap */
		r2 = error_norm;
        spmv_cpu<T>(h_spm, p, Ap);	
		alpha = r2 / Xdot(n, p, 1, Ap, 1);
        /* x := x + alpha p */
        Xaxpy(n, alpha, p, 1, h_x, 1);
        /* r := r - alpha Ap */
        Xaxpy(n, -alpha, Ap, 1, r, 1);
		/* beta := r'r / r2 */
        error_norm = Xnrm2(n, r, 1); 
		beta = error_norm / r2;
        /* p := r + beta p */
        Xscal(n, beta, p, 1);
        Xaxpy(n, 1.0, r, 1, p, 1);
        #if USE_ERROR_REL
        if(error_norm < error_norm_rel || k > max_iter)
            break;
        #else
        if(error_norm < ERROR_ABS || k > max_iter)
            break;        
        #endif
        if(!(k % SHOW_INTERVAL)) printf("HOST: interation num: %d, error: %f\n", k, error_norm);
	}
    stop_timer(timer_compu, &t_compu_ms);
    *iter = k;
    fprintf(fpDat, "The iteration of CPU is: %d\n", k);
    #if USE_ERROR_REL
    double Gflops = (t_compu_ms == 0.0) ? 0.0 : 
    ((2.0 * h_spm->nnzero + 5.0 * n + (k + 1) * (2.0 * h_spm->nnzero + 10.0 * n)) / t_compu_ms) / 1e6;
    #else
    double Gflops = (t_compu_ms == 0.0) ? 0.0 : 
    ((2.0 * h_spm->nnzero + 3.0 * n - 2.0 + (k + 1) * (2.0 * h_spm->nnzero + 10.0 * n)) / t_compu_ms) / 1e6;
    #endif
    fprintf(fpDat, "The compute GFLOPs for CG: %2.2f\n", Gflops);
    
	free(p);
	free(r);
	free(Ap);

	#if USE_ERROR_REL
	return sqrt(error_norm / b_square);
    #else
    return sqrt(error_norm / n);
    #endif
}

#endif /* __CG_HOST_H__ */
