/*
 * CUDA for Sparse Solver
 * Author: Luo Li     li.luo@sub.siat.ac.cn
 */ 

#ifndef __CG_MULTI_DEVICE_H__
#define __CG_MULTI_DEVICE_H__

#include <omp.h>
#include "cutils.h"

template <class Matrix_T, class T>
T cg_multi_device(FILE *fpDat, Matrix_T *h_spm, T *h_b, T *mh_x, const int max_iter, const T error_rel, int *iter)
{	
    int n;
    if(h_spm->nrow == h_spm->ncol) 
        n = h_spm->nrow;
    else
    {
        printf("CG_MULTI_DEVICE: Nonsquare matrix doesn't supported.\n");
        exit(EXIT_FAILURE);
    }

	int GPU_N, Thread_N;
	cutilSafeCall(cudaGetDeviceCount(&GPU_N));
	if(GPU_N > MAX_GPU_COUNT) 
        GPU_N = MAX_GPU_COUNT;
	fprintf(fpDat, "CUDA-capable device count: %i\n", GPU_N);
	Thread_N = GPU_PER_THREAD * GPU_N;

	/* handle the case when n / Thread_N != 0, all threads except the last one plus 1, 
	while the last one is quiter small, this approach is not good for balance, to achieve a good balance,
	will use the approach in petsc/ksp/examples/ex3.c */
	int n_per_thr, real_TN = 0;	
	n_per_thr = n / Thread_N;
	if (n % Thread_N != 0)
		n_per_thr++;
	while(n_per_thr * real_TN < n)
		real_TN++;

	T error_norm, pAp;	
    int max_nnz_per_thr = 0;
    int sum_nnz = 0;
    T *h_p = (T*)malloc(n * sizeof(T));

	omp_set_num_threads(real_TN);
    #if USE_ERROR_REL
    T b_square;
    #pragma omp parallel shared(error_norm, pAp, max_nnz_per_thr, sum_nnz, b_square, h_p) 
    #else
    #pragma omp parallel shared(error_norm, pAp, max_nnz_per_thr, sum_nnz, h_p) 
    #endif
{	    
	cublasStatus status;
	int thread_id;
	thread_id = omp_get_thread_num();
	cudaSetDevice(thread_id % GPU_N);

    unsigned int timer_compu = 0;
    T t_compu_ms = 0.0;
    unsigned int timer_commu_part = 0;
    T t_commu_part_ms = 0.0;
    T t_commu_total_ms = 0.0;    
    if(thread_id == 0)
	start_timer(&timer_compu);	

	/* initialize CUBLAS */
	status = cublasInit();
	if(status != CUBLAS_STATUS_SUCCESS)
	{
		fprintf(stderr, "!!!! CUBLAS initialization error\n");
		exit(EXIT_FAILURE);
	}	

	T *b_per_thr = 0;
	T *d_x = 0;	
	T *x_per_thr = 0;	
	T *p_per_thr = 0;
	T *r_per_thr = 0;
	T *p = 0;
	T *Ap_per_thr = 0;
	T *Ax_per_thr = 0;

    int thread_start = thread_id * n_per_thr; 
    n_per_thr = min(n_per_thr, n - thread_start);

    /* allocate and upload device memory */
    int nnzero_per_thr;
    Matrix_T md_spm = md_alloc_matrix<T>(h_spm, n_per_thr, thread_start, &nnzero_per_thr);
    #pragma omp critical(nnzero_per_thr) 
    {
        max_nnz_per_thr = max(max_nnz_per_thr, nnzero_per_thr);
        sum_nnz += nnzero_per_thr;
    }

    cublasAlloc(n_per_thr, sizeof(T), (void **)&b_per_thr);
    cublasAlloc(n_per_thr, sizeof(T), (void **)&x_per_thr);
	cublasAlloc(n, sizeof(T), (void **)&d_x);
    cublasSetVector(n_per_thr, sizeof(T), h_b + thread_start, 1, b_per_thr, 1);	
	cublasSetVector(n, sizeof(T), mh_x, 1, d_x, 1);	
    cublasXcopy(n_per_thr, d_x + thread_start, 1, x_per_thr, 1);
	
	/* allocate temporary vectors */
	cublasAlloc(n_per_thr, sizeof(T), (void **)&p_per_thr);
	cublasAlloc(n_per_thr, sizeof(T), (void **)&r_per_thr);
	cublasAlloc(n_per_thr, sizeof(T), (void **)&Ap_per_thr);
	cublasAlloc(n_per_thr, sizeof(T), (void **)&Ax_per_thr);
	cublasAlloc(n, sizeof(T), (void **)&p);	

    #if USE_ERROR_REL
    T b_square_per_thr = cublasXnrm2(n_per_thr, b_per_thr, 1);
    #pragma omp single nowait 
    b_square = 0.0;		
    #pragma omp critical(b_square) 
    b_square += b_square_per_thr;
    #pragma omp barrier
    T error_norm_rel = error_rel * error_rel * b_square;    
    #endif

	/* r := b - Ax */
    spmv_multi_gpu<T>(&md_spm, r_per_thr, d_x, n_per_thr);
    cublasXaxpy(n_per_thr, -1.0, b_per_thr, 1, r_per_thr, 1);
    cublasXscal(n_per_thr, -1.0, r_per_thr, 1);
	/* p := r */
	cublasXcopy(n_per_thr, r_per_thr, 1, p_per_thr, 1);

	T r2_per_thr = cublasXnrm2(n_per_thr, r_per_thr, 1);
	#pragma omp single nowait 
	error_norm = 0.0;		
	#pragma omp critical(error_norm1) 
	error_norm += r2_per_thr;
	#pragma omp barrier

    int k;
	T r2, alpha, beta, pAp_per_thr;
    #if SET_SUPER_MAX_ITER
    for(k = 0; k < SUPER_MAX_ITER; ++k)
    #else
    for(k = 0; k < n; ++k)
    #endif	
	{   
		/* alpha := r'r / p'Ap */
		r2 = error_norm;
		//transfer p_per_thr between neighbour GPU
		if(thread_id == 0)
            start_timer(&timer_commu_part);	
		cublasGetVector(n_per_thr, sizeof(T), p_per_thr, 1, h_p + thread_start, 1);
		#pragma omp barrier
        cublasSetVector(n, sizeof(T), h_p, 1, p, 1);
        if(thread_id == 0)
        {
            stop_timer(timer_commu_part, &t_commu_part_ms);
            t_commu_total_ms += t_commu_part_ms;
		}
		//matrix A multiple vector p
        spmv_multi_gpu<T>(&md_spm, Ap_per_thr, p, n_per_thr);
        pAp_per_thr = cublasXdot(n_per_thr, p_per_thr, 1, Ap_per_thr, 1);
		#pragma omp single nowait 
		pAp = 0.0;		
		#pragma omp critical(pAp) 
		pAp += pAp_per_thr;
		#pragma omp barrier
		alpha = r2 / pAp;
		/* x := x + alpha p */
		cublasXaxpy(n_per_thr, alpha, p_per_thr, 1, x_per_thr, 1);
		/* r := r - alpha Ap */
		cublasXaxpy(n_per_thr, -alpha, Ap_per_thr, 1, r_per_thr, 1);
		/* beta := r'r / r2 */
		r2_per_thr = cublasXnrm2(n_per_thr, r_per_thr, 1);
		#pragma omp single nowait 
		error_norm = 0.0;		
		#pragma omp critical(error_norm2) 
		error_norm += r2_per_thr;
		#pragma omp barrier
		beta = error_norm / r2;
		/* p := r + beta p */
		cublasXscal(n_per_thr, beta, p_per_thr, 1);
		cublasXaxpy(n_per_thr, 1.0, r_per_thr, 1, p_per_thr, 1);
        #if USE_ERROR_REL
		if(error_norm < error_norm_rel || k > max_iter)
			break;
        #else
        if(error_norm < ERROR_ABS || k > max_iter)
			break;
        #endif
        if(!(k % SHOW_INTERVAL))
        {
            if(thread_id == 0) 
                printf("MULTI_DEVICE: interation num: %d, error: %f\n", k, error_norm);
        }
	}
	if(thread_id == 0)
    {
        stop_timer(timer_compu, &t_compu_ms);
        *iter = k;
        fprintf(fpDat, "The communication time of exchanging data is: %f ms\n", t_commu_total_ms);
        fprintf(fpDat, "The computation time of Multi-GPU is: %f ms\n", t_compu_ms - t_commu_total_ms);
        fprintf(fpDat, "The iteration of multi-GPU is: %d\n", k);

        #if USE_ERROR_REL
        T Gflops = (t_compu_ms == 0.0) ? 0.0 : 
        ((2.0 * h_spm->nnzero + 7.0 * n  + (k + 1) * (2.0 * h_spm->nnzero + 10.0 * n)) / t_compu_ms) / 1e6;
        #else
        T Gflops = (t_compu_ms == 0.0) ? 0.0 : 
        ((2.0 * h_spm->nnzero + 5.0 * n - 1.0 + (k + 1) * (2.0 * h_spm->nnzero + 10.0 * n)) / t_compu_ms) / 1e6;
        #endif
        fprintf(fpDat, "The compute GFLOPs for CG: %2.2f\n", Gflops);
	}
	/* transfer solution from device to host */
	cublasGetVector(n_per_thr, sizeof(T), x_per_thr, 1, mh_x + thread_start, 1);
    md_dealloc_matrix<T>(&md_spm);

	/* free temporary vectors */	
	cublasFree(p);
	cublasFree(p_per_thr);
	cublasFree(r_per_thr);
	cublasFree(Ap_per_thr);
	cublasFree(Ax_per_thr);
	/* free device memory */ 
	cublasFree(b_per_thr);
	cublasFree(d_x);
	cublasFree(x_per_thr);
    
	/* shutdown CUBLAS */
	status = cublasShutdown();
	if(status != CUBLAS_STATUS_SUCCESS)
	{
		fprintf(stderr, "!!!! shutdown error\n");
		exit(EXIT_FAILURE);
	}
}
    T imbalance = sum_nnz / T(max_nnz_per_thr * real_TN);
    fprintf(fpDat, "The imbalance of multi-GPU is: %2.2f\n", imbalance);  
    
    free(h_p);	

    #if USE_ERROR_REL
    return sqrt(error_norm / b_square);
    #else
    return sqrt(error_norm / n);
    #endif
}

#endif /* __CG_MULTI_DEVICE_H__ */
