/*
 * CUDA for Sparse Solver
 * Author: Luo Li     li.luo@sub.siat.ac.cn
 */ 

#ifndef __CSR_FORMAT_H__
#define __CSR_FORMAT_H__

#include "blas.h"

template <class T>
struct csr_format
{
    int nrow;
    int ncol;
    int nnzero;
    int interval_size;
    int *row_ptr;    
    int *col_ind;
    T *nz_val;
};

template <class T>
int cmp_csr(const void *e1, const void *e2) 
{
    if (((nzinfo<T> *)e1)->row_num < ((nzinfo<T> *)e2)->row_num) return -1;
    if (((nzinfo<T> *)e1)->row_num > ((nzinfo<T> *)e2)->row_num) return 1;
    if (((nzinfo<T> *)e1)->col_num < ((nzinfo<T> *)e2)->col_num) return -1;
    if (((nzinfo<T> *)e1)->col_num > ((nzinfo<T> *)e2)->col_num) return -1;
    return 0;
}

template <class T>
void convert_matrix(spmatrix<T> *spm, csr_format<T> *spm_csr)
{
    qsort(spm->nzentry, spm->nnzero, sizeof(nzinfo<T>), cmp_csr<T>);

    spm_csr->nrow = spm->nrow; 
    spm_csr->ncol = spm->ncol; 
    spm_csr->nnzero = spm->nnzero;

    #if USE_ALIGNED_OPT_SPM
    spm_csr->row_ptr = (int *)malloc(sizeof(int) * (spm_csr->nrow + 1));

    int currow, prevrow = (spm->nzentry)[0].row_num;
    for(int tmp_row = 0; tmp_row <= prevrow; tmp_row++)
        (spm_csr->row_ptr)[tmp_row] = 0;    
    
    for(int i = 1; i < spm_csr->nnzero; i++) 
    {
        currow = (spm->nzentry)[i].row_num;
        if(currow != prevrow && currow != (prevrow + 1)) 
        {
            for(int tmp_row = prevrow + 1; tmp_row <= currow; tmp_row++)
                (spm_csr->row_ptr)[tmp_row] = i;
            prevrow = currow;
        }
        else if(currow == (prevrow + 1))
        {
            prevrow = currow;
            (spm_csr->row_ptr)[currow] = i;
        }
    }
    if(spm_csr->nrow != (prevrow + 1))
    {
        for(int tmp_row = prevrow + 1; tmp_row < spm_csr->nrow; tmp_row++)
            (spm_csr->row_ptr)[tmp_row] = spm_csr->nnzero;        
    }
    (spm_csr->row_ptr)[spm_csr->nrow] = spm_csr->nnzero;    
    
    int nnz_row, pad_nnzero = 0;
    for(int i = 0; i < spm_csr->nrow; i++) 
    {
        nnz_row = (spm_csr->row_ptr)[i + 1] - (spm_csr->row_ptr)[i];
        if(nnz_row % HALFWARP) 
            pad_nnzero += nnz_row + HALFWARP - (nnz_row % HALFWARP); 
        else 
            pad_nnzero += nnz_row;
    }

    free(spm_csr->row_ptr);
    #if USE_PAGE_LOCKED
    CUDA_SAFE_CALL(cudaMallocHost((void **)&spm_csr->row_ptr, (spm_csr->nrow + 1) * sizeof(int)));
    CUDA_SAFE_CALL(cudaMallocHost((void **)&spm_csr->col_ind, pad_nnzero * sizeof(int)));
    CUDA_SAFE_CALL(cudaMallocHost((void **)&spm_csr->nz_val, pad_nnzero * sizeof(T)));
    #else
    spm_csr->row_ptr = (int *)malloc(sizeof(int) * (spm_csr->nrow + 1));
    spm_csr->col_ind = (int *)malloc(sizeof(int) * pad_nnzero);
    spm_csr->nz_val = (T *)malloc(sizeof(T) * pad_nnzero);
    #endif

    int cur_row, prev_row = (spm->nzentry)[0].row_num;
    for(int tmp_row = 0; tmp_row <= prev_row; tmp_row++)
        (spm_csr->row_ptr)[tmp_row] = 0;    

    int cur_pad_nnzero = 0; 
    for(int i = 0; i < spm_csr->nnzero; i++) 
    {
        cur_row = (spm->nzentry)[i].row_num;
        int cur_mode = cur_pad_nnzero % HALFWARP;
        if(cur_row != prev_row && cur_row != (prev_row + 1)) 
        {
            if(cur_mode)
            {
                int fillCount = HALFWARP - cur_mode;
                for(int j = 0; j < fillCount; j++) 
                {
                    (spm_csr->col_ind)[cur_pad_nnzero] = -1;
                    (spm_csr->nz_val)[cur_pad_nnzero] = 0.0;
                    cur_pad_nnzero++;
                }
            }
            for(int tmp_row = prev_row + 1; tmp_row <= cur_row; tmp_row++)
                (spm_csr->row_ptr)[tmp_row] = cur_pad_nnzero;
            prev_row = cur_row;
        }
        else if(cur_row == (prev_row + 1))
        {
            if(cur_mode)
            {
                int fillCount = HALFWARP - cur_mode;
                for(int j = 0; j < fillCount; j++) 
                {
                    (spm_csr->col_ind)[cur_pad_nnzero] = -1;
                    (spm_csr->nz_val)[cur_pad_nnzero] = 0.0;
                    cur_pad_nnzero++;
                }
            }            
            prev_row = cur_row;
            (spm_csr->row_ptr)[cur_row] = cur_pad_nnzero;
        }
        (spm_csr->col_ind)[cur_pad_nnzero] = (spm->nzentry)[i].col_num;
        (spm_csr->nz_val)[cur_pad_nnzero] = (spm->nzentry)[i].val;        
        cur_pad_nnzero++;
    }
    int cur_mode = cur_pad_nnzero % HALFWARP;
    if(cur_mode)
    {
        int fillCount = HALFWARP - cur_mode;
        for(int j = 0; j < fillCount; j++) 
        {
            (spm_csr->col_ind)[cur_pad_nnzero] = -1;
            (spm_csr->nz_val)[cur_pad_nnzero] = 0.0;
            cur_pad_nnzero++;
        }
    }    
    if(spm_csr->nrow != (prev_row + 1))
    {
        for(int tmp_row = prev_row + 1; tmp_row < spm_csr->nrow; tmp_row++)
            (spm_csr->row_ptr)[tmp_row] = pad_nnzero;        
    }
    (spm_csr->row_ptr)[spm_csr->nrow] = pad_nnzero;
    spm_csr->nnzero = pad_nnzero;
    #else
    #if USE_PAGE_LOCKED
    CUDA_SAFE_CALL(cudaMallocHost((void **)&spm_csr->row_ptr, (spm_csr->nrow + 1) * sizeof(int)));
    CUDA_SAFE_CALL(cudaMallocHost((void **)&spm_csr->col_ind, spm_csr->nnzero * sizeof(int)));
    CUDA_SAFE_CALL(cudaMallocHost((void **)&spm_csr->nz_val, spm_csr->nnzero * sizeof(T)));
    #else
    spm_csr->row_ptr = (int *)malloc(sizeof(int) * (spm_csr->nrow + 1));
    spm_csr->col_ind = (int *)malloc(sizeof(int) * spm_csr->nnzero);
    spm_csr->nz_val = (T *)malloc(sizeof(T) * spm_csr->nnzero);
    #endif

    int cur_row, prev_row = (spm->nzentry)[0].row_num;
    for(int tmp_row = 0; tmp_row <= prev_row; tmp_row++)
        (spm_csr->row_ptr)[tmp_row] = 0;    
    
    for(int i = 1; i < spm_csr->nnzero; i++) 
    {
        cur_row = (spm->nzentry)[i].row_num;
        if(cur_row != prev_row && cur_row != (prev_row + 1)) 
        {
            for(int tmp_row = prev_row + 1; tmp_row <= cur_row; tmp_row++)
                (spm_csr->row_ptr)[tmp_row] = i;
            prev_row = cur_row;
        }
        else if(cur_row == (prev_row + 1))
        {
            prev_row = cur_row;
            (spm_csr->row_ptr)[cur_row] = i;
        }
    }
    if(spm_csr->nrow != (prev_row + 1))
    {
        for(int tmp_row = prev_row + 1; tmp_row < spm_csr->nrow; tmp_row++)
            (spm_csr->row_ptr)[tmp_row] = spm_csr->nnzero;        
    }
    (spm_csr->row_ptr)[spm_csr->nrow] = spm_csr->nnzero;
    
    for(int i = 0; i < spm_csr->nnzero; i++) 
    {
        (spm_csr->col_ind)[i] = (spm->nzentry)[i].col_num;
        (spm_csr->nz_val)[i] = (spm->nzentry)[i].val;
    }    
    #endif

    /* automatically alloc thread number to spm_csr->interval_size */
    #if USE_DYNAMIC_INTERVAL_SIZE
    //even num that larger than average size
    spm_csr->interval_size = spm_csr->nnzero / spm_csr->nrow + 1;
    if(spm_csr->interval_size > MAX_INTERVAL_SIZE)
        spm_csr->interval_size = MAX_INTERVAL_SIZE;
    else
    {
        for(int i = 1; i <= MAX_INTERVAL_SIZE; i *= 2)
        if(spm_csr->interval_size / i == 0)
        {
            spm_csr->interval_size = i;
            break;
        }
    }
    #else
    //fixed number
    spm_csr->interval_size = INTERVAL_SIZE;
    #endif    
}

template <class T>
void print_matrix(csr_format<T> *spm_csr)
{
    printf("CSR format: \n");
    for(int i = 0; i < spm_csr->nrow; i++) 
    {
        for(int jj = (spm_csr->row_ptr)[i]; jj < (spm_csr->row_ptr)[i + 1]; jj++) 
            printf("%f ", (spm_csr->nz_val)[jj]);
        printf("\n");
    }
    printf("\n\n");
}

template <class T>
void clear_format(csr_format<T> *spm_csr) 
{
    #if USE_PAGE_LOCKED
    CUDA_SAFE_CALL(cudaFreeHost(spm_csr->row_ptr));
    CUDA_SAFE_CALL(cudaFreeHost(spm_csr->col_ind));
    CUDA_SAFE_CALL(cudaFreeHost(spm_csr->nz_val));
    #else
    free(spm_csr->row_ptr);
    free(spm_csr->col_ind);
    free(spm_csr->nz_val);
    #endif
}

template <class T>
csr_format<T> d_alloc_matrix(csr_format<T> *h_spm)
{   
    csr_format<T> d_spm;
    d_spm.nrow = h_spm->nrow;
    d_spm.ncol = h_spm->ncol;
    d_spm.nnzero = h_spm->nnzero;
    d_spm.interval_size = h_spm->interval_size;

    CUDA_SAFE_CALL(cudaMalloc((void **)&d_spm.row_ptr, (d_spm.nrow + 1) * sizeof(int)));
    CUDA_SAFE_CALL(cudaMalloc((void **)&d_spm.col_ind, d_spm.nnzero * sizeof(int)));
    CUDA_SAFE_CALL(cudaMalloc((void **)&d_spm.nz_val, d_spm.nnzero * sizeof(T)));

    CUDA_SAFE_CALL(cudaMemcpy(d_spm.row_ptr, h_spm->row_ptr, (d_spm.nrow + 1) * sizeof(int), cudaMemcpyHostToDevice));
    CUDA_SAFE_CALL(cudaMemcpy(d_spm.col_ind, h_spm->col_ind, d_spm.nnzero * sizeof(int), cudaMemcpyHostToDevice));
    CUDA_SAFE_CALL(cudaMemcpy(d_spm.nz_val, h_spm->nz_val, d_spm.nnzero * sizeof(T), cudaMemcpyHostToDevice));
    
    return d_spm;
}

template <class T>
void d_dealloc_matrix(csr_format<T> *d_spm)
{
    CUDA_SAFE_CALL(cudaFree(d_spm->row_ptr));
    CUDA_SAFE_CALL(cudaFree(d_spm->col_ind));
    CUDA_SAFE_CALL(cudaFree(d_spm->nz_val));
}

template <class T>
csr_format<T> md_alloc_matrix(csr_format<T> *h_spm, const int n_per_thr, const int thread_start, int *nnzero_per_thr)
{   
    csr_format<T> md_spm;
    md_spm.nrow = h_spm->nrow;
    md_spm.ncol = h_spm->ncol;
    md_spm.nnzero = h_spm->nnzero;
    md_spm.interval_size = h_spm->interval_size;

    cublasAlloc(n_per_thr + 1, sizeof(int), (void **)&md_spm.row_ptr);
    int vnum_per_thr = (h_spm->row_ptr)[thread_start + n_per_thr] - (h_spm->row_ptr)[thread_start];
    *nnzero_per_thr = vnum_per_thr;
    cublasAlloc(vnum_per_thr, sizeof(int), (void **)&md_spm.col_ind);
    cublasAlloc(vnum_per_thr, sizeof(T), (void **)&md_spm.nz_val);

    cublasSetVector(n_per_thr + 1, sizeof(int), h_spm->row_ptr + thread_start, 1, md_spm.row_ptr, 1);
    cublasSetVector(vnum_per_thr, sizeof(int), h_spm->col_ind + (h_spm->row_ptr)[thread_start], 1, md_spm.col_ind, 1);
    cublasSetVector(vnum_per_thr, sizeof(T), h_spm->nz_val + (h_spm->row_ptr)[thread_start], 1, md_spm.nz_val, 1);
    
    return md_spm;
}

template <class T>
void md_dealloc_matrix(csr_format<T> *md_spm)
{
    cublasFree(md_spm->row_ptr);
    cublasFree(md_spm->col_ind);
    cublasFree(md_spm->nz_val);
}

template <class T>
void spmv_csr_cpu(int *h_row, int *h_col, T *h_val, T *h_vec, T *h_result_vec, const int m)
{
	for(int i = 0; i < m; i++)
    {
        T sum = 0.0;
		for(int j = h_row[i]; j < h_row[i + 1]; j++)
			 sum +=  h_val[j] * h_vec[h_col[j]];
        h_result_vec[i] = sum;
    }
}

template <class T>
void spmv_cpu(csr_format<T> *h_spm, T *h_vec, T *h_result_vec)
{
    int m = h_spm->nrow;
    spmv_csr_cpu<T>(h_spm->row_ptr, h_spm->col_ind, h_spm->nz_val, h_vec, h_result_vec, m);
}

template <class T>
void spmv_single_gpu(csr_format<T> *d_spm, T *d_result_vec, T *d_vec)
{
    int n = d_spm->ncol;
    int interval_size = d_spm->interval_size;	
    unsigned int dimBlock = DIMBLOCK;
    unsigned int rows_per_block = dimBlock / interval_size;	
    const dim3 dimGrid = large_grid(n, rows_per_block);
    unsigned int mem_size = dimBlock * sizeof(T);

    #if USE_TEXTURE
    bind_vec(d_vec);
    #endif
    spmv_csr_single_gpu<T><<<dimGrid, dimBlock, mem_size>>>(d_spm->row_ptr, d_spm->col_ind, d_spm->nz_val, d_result_vec, d_vec, interval_size, n);
    #if USE_TEXTURE
    unbind_vec(d_vec);
    #endif    
}

template <class T>
void spmv_multi_gpu(csr_format<T> *md_spm, T *d_result_vec, T *d_vec, const int n_per_thr)
{
    int interval_size = md_spm->interval_size;
    unsigned int dimBlock = DIMBLOCK;
    unsigned int rows_per_block = dimBlock / interval_size;
    const dim3 dimGrid = large_grid(n_per_thr, rows_per_block);
    unsigned int mem_size = dimBlock * sizeof(double);	

    #if USE_TEXTURE
    bind_vec(d_vec);
    #endif    
    spmv_csr_multi_gpu<T><<<dimGrid, dimBlock, mem_size>>>(md_spm->row_ptr, md_spm->col_ind, md_spm->nz_val, d_result_vec, d_vec, interval_size, n_per_thr);
    #if USE_TEXTURE
    unbind_vec(d_vec);
    #endif    
}

template <class T>
void preconditioner_Jacobi_csr_cpu(int *h_row, int *h_col, T *h_val, T *r, T *h, const int n)
{
    for(int i = 0; i < n; i++)
    {   
        T r_val = r[i];
        for(int j = h_row[i]; j < h_row[i + 1]; j++)
            if(i == h_col[j]) 
            {   
                T val_temp = h_val[j];
                T val = (val_temp == 0.0) ? 1.0 : 1.0 / val_temp;
                r_val = r_val * val;
            }
        h[i] = r_val;
    }
}

template <class T>
void precon_ssor_Lower_csr_cpu(int *h_row, int *h_col, T *h_val, T *x, T *y, const int n, const T omega)
{
    for(int i = 0; i < n; i++)
    {
        T sum = 0.0;
        T diag;
        for(int j = h_row[i]; j < h_row[i + 1]; j++)
        {
            int col_num = h_col[j];
            T val = h_val[j];
            if(col_num < i)
                sum -= val * y[col_num];
            if(col_num == i)
                diag = val;
        }
        y[i] = (x[i] - sum) * omega / diag;//(D/w - L) rather than L !!! so use negative
    }    
}

template <class T>
void precon_ssor_Diag_csr_cpu(int *h_row, int *h_col, T *h_val, T *x, const int n, const T omega)
{
    for(int i = 0; i < n; i++)
    {   
        for(int j = h_row[i]; j < h_row[i + 1]; j++)
            if(i == h_col[j]) 
                x[i] *= h_val[j] / omega;
    }    
}

template <class T>
void precon_ssor_Upper_csr_cpu(int *h_row, int *h_col, T *h_val, T *x, T *y, const int n, const T omega)
{
    for(int i = n - 1; i >= 0; i--)
    {
        T sum = 0.0;
        T diag;
        for(int j = h_row[i]; j < h_row[i + 1]; j++)
        {
            int col_num = h_col[j];
            T val = h_val[j];
            if(col_num > i)
                sum -= val * y[col_num];
            if(col_num == i)
                diag = val;
        }
        y[i] = (x[i] - sum) * omega / diag;
    }     
}

template <class T>
void preconditioner_ssor_csr_cpu(int *h_row, int *h_col, T *h_val, T *r, T *h, const int n)
{
    T omega = OMEGA;
    T *temp = (T *)malloc(sizeof(T) * n);
    precon_ssor_Lower_csr_cpu<T>(h_row, h_col, h_val, r, temp, n, omega);
    precon_ssor_Diag_csr_cpu<T>(h_row, h_col, h_val, temp, n, omega);
    precon_ssor_Upper_csr_cpu<T>(h_row, h_col, h_val, temp, h, n, omega);
    sca_mul_vec<T>(n, 2.0 - omega, h, h);
    free(temp);
}

template <class T>
void preconditioner_cpu(char *precon_name, csr_format<T> *h_spm, T *r, T *h)
{
    int n = h_spm->ncol;
    #if USE_FLOAT_PRECON
    int nnzero = h_spm->nnzero;
    float *h_val = (float *)malloc(nnzero * sizeof(float));
    float *r_float = (float *)malloc(n * sizeof(float));
    float *h_float = (float *)malloc(n * sizeof(float));
    std::transform(h_spm->nz_val, h_spm->nz_val + nnzero, h_val, convert<T, float>());
    std::transform(r, r + n, r_float, convert<T, float>());
    std::transform(h, h + n, h_float, convert<T, float>());
    if(!strcmp(precon_name, "Jacobi"))
    preconditioner_Jacobi_csr_cpu<float>(h_spm->row_ptr, h_spm->col_ind, h_val, r_float, h_float, n);
    else if(!strcmp(precon_name, "SSOR"))
    preconditioner_ssor_csr_cpu<float>(h_spm->row_ptr, h_spm->col_ind, h_val, r_float, h_float, n);
    std::transform(h_float, h_float + n, h, convert<float, T>());
    free(h_val);
    free(r_float);
    free(h_float);
    #else
    if(!strcmp(precon_name, "Jacobi"))
    preconditioner_Jacobi_csr_cpu<T>(h_spm->row_ptr, h_spm->col_ind, h_spm->nz_val, r, h, n);
    else if(!strcmp(precon_name, "SSOR"))
    preconditioner_ssor_csr_cpu<T>(h_spm->row_ptr, h_spm->col_ind, h_spm->nz_val, r, h, n);
    #endif
} 

template <class T>
void preconditioner_single_gpu(char *precon_name, csr_format<T> *d_spm, T *r, T *h)
{
    int n = d_spm->ncol;
    //int interval_size = d_spm->interval_size;	
    unsigned int dimBlock = DIMBLOCK;
    //unsigned int rows_per_block = dimBlock / interval_size;
    //const dim3 dimGrid_vec = large_grid(n, rows_per_block);
    const dim3 dimGrid_scal = large_grid(n, dimBlock); 
    
    #if USE_FLOAT_PRECON
    int nnzero = d_spm->nnzero;
    float *d_val = 0;
    float *r_float = 0;
    float *h_float = 0;
    CUDA_SAFE_CALL(cudaMalloc((void **)&d_val, nnzero * sizeof(float)));
    CUDA_SAFE_CALL(cudaMalloc((void **)&r_float, n * sizeof(float)));
    CUDA_SAFE_CALL(cudaMalloc((void **)&h_float, n * sizeof(float)));
    const dim3 dimGrid_val = large_grid(nnzero, dimBlock);
    transform_gpu<T, float><<<dimGrid_val, dimBlock>>>(d_spm->nz_val, d_val, nnzero);
    const dim3 dimGrid_col = large_grid(n, dimBlock);
    transform_gpu<T, float><<<dimGrid_col, dimBlock>>>(r, r_float, n);
    transform_gpu<T, float><<<dimGrid_col, dimBlock>>>(h, h_float, n);
    #if USE_TEXTURE
    bind_vec(r_float);
    #endif 
    if(!strcmp(precon_name, "Jacobi"))
    {
        //preconditioner_Jacobi_vec_csr_single_gpu<float><<<dimGrid_vec, dimBlock>>>(d_spm->row_ptr, d_spm->col_ind, d_val, r_float, h_float, interval_size, n);
        preconditioner_Jacobi_scal_csr_single_gpu<float><<<dimGrid_scal, dimBlock>>>(d_spm->row_ptr, d_spm->col_ind, d_val, r_float, h_float, n);    
    }else if(!strcmp(precon_name, "SSOR"))
        {
            preconditioner_Jacobi_scal_csr_single_gpu<float><<<dimGrid_scal, dimBlock>>>(d_spm->row_ptr, d_spm->col_ind, d_val, r_float, h_float, n);    
        }
    #if USE_TEXTURE
    unbind_vec(r_float);
    #endif 
    transform_gpu<float, T><<<dimGrid_col, dimBlock>>>(h_float, h, n);
    cublasFree(d_val);
    cublasFree(r_float);
    cublasFree(h_float);
    #else
    #if USE_TEXTURE
    bind_vec(r);
    #endif
    if(!strcmp(precon_name, "Jacobi"))
    {     
        //preconditioner_Jacobi_vec_csr_single_gpu<T><<<dimGrid_vec, dimBlock>>>(d_spm->row_ptr, d_spm->col_ind, d_spm->nz_val, r, h, interval_size, n);
        preconditioner_Jacobi_scal_csr_single_gpu<T><<<dimGrid_scal, dimBlock>>>(d_spm->row_ptr, d_spm->col_ind, d_spm->nz_val, r, h, n);
    }else if(!strcmp(precon_name, "SSOR"))
        {
            preconditioner_Jacobi_scal_csr_single_gpu<T><<<dimGrid_scal, dimBlock>>>(d_spm->row_ptr, d_spm->col_ind, d_spm->nz_val, r, h, n);
        }
    #if USE_TEXTURE
    unbind_vec(r);
    #endif     
    #endif        
}

template <class T>
void preconditioner_multi_gpu(char *precon_name, csr_format<T> *md_spm, T *r, T *h, const int n_per_thr, const int nnzero_per_thr, const int thread_start)
{
    //int interval_size = md_spm->interval_size;
    unsigned int dimBlock = DIMBLOCK;
    //unsigned int rows_per_block = dimBlock / interval_size;
    //const dim3 dimGrid_vec = large_grid(n_per_thr, rows_per_block);
    const dim3 dimGrid_scal = large_grid(n_per_thr, dimBlock);    
    
    #if USE_FLOAT_PRECON
    float *md_val = 0;
    float *r_float = 0;
    float *h_float = 0;
    CUDA_SAFE_CALL(cudaMalloc((void **)&md_val, nnzero_per_thr * sizeof(float)));
    CUDA_SAFE_CALL(cudaMalloc((void **)&r_float, n_per_thr * sizeof(float)));
    CUDA_SAFE_CALL(cudaMalloc((void **)&h_float, n_per_thr * sizeof(float)));
    const dim3 dimGrid_val = large_grid(nnzero_per_thr, dimBlock);
    transform_gpu<T, float><<<dimGrid_val, dimBlock>>>(md_spm->nz_val, md_val, nnzero_per_thr);
    const dim3 dimGrid_col = large_grid(n_per_thr, dimBlock);
    transform_gpu<T, float><<<dimGrid_col, dimBlock>>>(r, r_float, n_per_thr);
    transform_gpu<T, float><<<dimGrid_col, dimBlock>>>(h, h_float, n_per_thr);
    #if USE_TEXTURE
    bind_vec(r_float);
    #endif    
    if(!strcmp(precon_name, "Jacobi"))
    {     
        //preconditioner_Jacobi_vec_csr_multi_gpu<float><<<dimGrid_vec, dimBlock>>>(md_spm->row_ptr, md_spm->col_ind, md_val, r_float, h_float, interval_size, n_per_thr, thread_start);
        preconditioner_Jacobi_scal_csr_multi_gpu<float><<<dimGrid_scal, dimBlock>>>(md_spm->row_ptr, md_spm->col_ind, md_val, r_float, h_float, n_per_thr, thread_start);    
    }else if(!strcmp(precon_name, "SSOR"))
        {
            //preconditioner_Jacobi_vec_csr_multi_gpu<float><<<dimGrid_vec, dimBlock>>>(md_spm->row_ptr, md_spm->col_ind, md_val, r_float, h_float, interval_size, n_per_thr, thread_start);
            preconditioner_Jacobi_scal_csr_multi_gpu<float><<<dimGrid_scal, dimBlock>>>(md_spm->row_ptr, md_spm->col_ind, md_val, r_float, h_float, n_per_thr, thread_start);            
        }
    #if USE_TEXTURE
    unbind_vec(r_float);
    #endif     
    transform_gpu<float, T><<<dimGrid_col, dimBlock>>>(h_float, h, n_per_thr);
    cublasFree(md_val);
    cublasFree(r_float);
    cublasFree(h_float);
    #else
    #if USE_TEXTURE
    bind_vec(r);
    #endif    
    if(!strcmp(precon_name, "Jacobi"))
    {     
        //preconditioner_Jacobi_vec_csr_multi_gpu<T><<<dimGrid_vec, dimBlock>>>(md_spm->row_ptr, md_spm->col_ind, md_spm->nz_val, r, h, interval_size, n_per_thr, thread_start);
        preconditioner_Jacobi_scal_csr_multi_gpu<T><<<dimGrid_scal, dimBlock>>>(md_spm->row_ptr, md_spm->col_ind, md_spm->nz_val, r, h, n_per_thr, thread_start);    
    }else if(!strcmp(precon_name, "SSOR"))
        {
            //preconditioner_Jacobi_vec_csr_multi_gpu<T><<<dimGrid_vec, dimBlock>>>(md_spm->row_ptr, md_spm->col_ind, md_spm->nz_val, r, h, interval_size, n_per_thr, thread_start);
            preconditioner_Jacobi_scal_csr_multi_gpu<T><<<dimGrid_scal, dimBlock>>>(md_spm->row_ptr, md_spm->col_ind, md_spm->nz_val, r, h, n_per_thr, thread_start);    
        }        
    #if USE_TEXTURE
    unbind_vec(r);
    #endif     
    #endif        
}

#endif /* __CSR_FORMAT_H__ */
