/*
 * CUDA for Sparse Solver
 * Author: Luo Li     li.luo@sub.siat.ac.cn
 */ 

#ifndef __ELL_FORMAT_H__
#define __ELL_FORMAT_H__

#include "blas.h"

template <class T>
struct ell_format
{
    int nrow;
    int ncol;
    int nnzero;
    int ncol_per_row;
    int *indices;
    T *nz_val;
    #if USE_ALIGNED_OPT_SPM
    int stride;
    #endif
};

template <class T>
void convert_matrix(spmatrix<T> *spm, ell_format<T> *spm_ell)
{
    spm_ell->nrow = spm->nrow; 
    spm_ell->ncol = spm->ncol; 
    spm_ell->nnzero = spm->nnzero;
    
    int max_nnz_per_row = 0;
    for(int i = 0; i < spm_ell->nrow; i++) 
        max_nnz_per_row = max(max_nnz_per_row, (spm->nnz_per_row)[i]);    
    spm_ell->ncol_per_row = max_nnz_per_row;

    #if USE_ALIGNED_OPT_SPM
    spm_ell->stride = HALFWARP * ((spm_ell->nrow + HALFWARP - 1)/ HALFWARP);
    int ell_size = spm_ell->stride * spm_ell->ncol_per_row;
    #else
    int ell_size = spm_ell->nrow * spm_ell->ncol_per_row;
    #endif

    #if USE_PAGE_LOCKED
    CUDA_SAFE_CALL(cudaMallocHost((void **)&spm_ell->indices, ell_size * sizeof(int)));
    CUDA_SAFE_CALL(cudaMallocHost((void **)&spm_ell->nz_val, ell_size * sizeof(T)));
    #else
    spm_ell->indices = (int *)malloc(sizeof(int) * ell_size);
    spm_ell->nz_val = (T *)malloc(sizeof(T) * ell_size);
    #endif

    memset(spm_ell->indices, -1, ell_size * sizeof(int));
    memset(spm_ell->nz_val, 0, ell_size * sizeof(T));

    int *nnz_per_row_tmp = (int *)malloc(sizeof(int) * spm_ell->nrow);
    memset(nnz_per_row_tmp, 0, spm_ell->nrow * sizeof(int));
    
    for (int i = 0; i < spm->nnzero; i++) 
    {
        int ii = (spm->nzentry)[i].row_num;//current row_num in ELL format
        int jj = nnz_per_row_tmp[ii];//current col_num in ELL format
        nnz_per_row_tmp[ii]++;
        #if USE_ALIGNED_OPT_SPM
        int index = spm_ell->stride * jj + ii;
        #else
        int index = spm_ell->nrow * jj + ii;
        #endif        
        (spm_ell->indices)[index] = (spm->nzentry)[i].col_num;
        (spm_ell->nz_val)[index] = (spm->nzentry)[i].val;
    }   
    free(nnz_per_row_tmp);
}

template <class T>
void print_matrix(ell_format<T> *spm_ell)
{
    printf("ELL format: \n");
    #if USE_ALIGNED_OPT_SPM
    for(int ii = 0; ii < spm_ell->stride; ii++) 
    #else
    for(int ii = 0; ii < spm_ell->nrow; ii++) 
    #endif
    {
        for(int jj = 0; jj < spm_ell->ncol_per_row; jj++) 
        {
            #if USE_ALIGNED_OPT_SPM
            int index = spm_ell->stride * jj + ii;
            #else
            int index = spm_ell->nrow * jj + ii;
            #endif
            printf("%f ", (spm_ell->nz_val)[index]);
        }
        printf("\n");
    }
    printf("\n\n");
}

template <class T>
void clear_format(ell_format<T> *spm_ell) 
{
    #if USE_PAGE_LOCKED
    CUDA_SAFE_CALL(cudaFreeHost(spm_ell->indices));
    CUDA_SAFE_CALL(cudaFreeHost(spm_ell->nz_val));
    #else
    free(spm_ell->indices);
    free(spm_ell->nz_val);
    #endif
}

template <class T>
ell_format<T> d_alloc_matrix(ell_format<T> *h_spm)
{  
    ell_format<T> d_spm;
    d_spm.nrow = h_spm->nrow;
    d_spm.ncol = h_spm->ncol;
    d_spm.nnzero = h_spm->nnzero;
    d_spm.ncol_per_row = h_spm->ncol_per_row;   
    #if USE_ALIGNED_OPT_SPM
    d_spm.stride = h_spm->stride;
    int ell_size = d_spm.stride * d_spm.ncol_per_row;
    #else
    int ell_size = d_spm.nrow * d_spm.ncol_per_row;
    #endif
    
    CUDA_SAFE_CALL(cudaMalloc((void **)&d_spm.indices, ell_size * sizeof(int)));
    CUDA_SAFE_CALL(cudaMalloc((void **)&d_spm.nz_val, ell_size * sizeof(T)));
    
    CUDA_SAFE_CALL(cudaMemcpy(d_spm.indices, h_spm->indices, ell_size * sizeof(int), cudaMemcpyHostToDevice));
    CUDA_SAFE_CALL(cudaMemcpy(d_spm.nz_val, h_spm->nz_val, ell_size * sizeof(T), cudaMemcpyHostToDevice));
    
    return d_spm;    
}

template <class T>
void d_dealloc_matrix(ell_format<T> *d_spm)
{
    CUDA_SAFE_CALL(cudaFree(d_spm->indices));
    CUDA_SAFE_CALL(cudaFree(d_spm->nz_val));
}

template <class T>
ell_format<T> md_alloc_matrix(ell_format<T> *h_spm, const int n_per_thr, const int thread_start, int *nnzero_per_thr)
{   
    ell_format<T> md_spm;
    md_spm.nrow = h_spm->nrow;
    md_spm.ncol = h_spm->ncol;
    md_spm.nnzero = h_spm->nnzero;
    md_spm.ncol_per_row = h_spm->ncol_per_row; 
    #if USE_ALIGNED_OPT_SPM
    md_spm.stride = HALFWARP * ((n_per_thr + HALFWARP - 1)/ HALFWARP);
    int ell_size = md_spm.stride * md_spm.ncol_per_row;
    #else
    int ell_size = n_per_thr * md_spm.ncol_per_row;
    #endif
    *nnzero_per_thr = ell_size;

    CUDA_SAFE_CALL(cudaMalloc((void **)&md_spm.indices, ell_size * sizeof(int)));
    CUDA_SAFE_CALL(cudaMalloc((void **)&md_spm.nz_val, ell_size * sizeof(T)));

    #if USE_ALIGNED_OPT_SPM
    int *tmp_indices = (int *)malloc(sizeof(int) * ell_size);
    T *tmp_nz_val = (T *)malloc(sizeof(T) * ell_size);
    memset(tmp_indices, 0, ell_size * sizeof(int));
    memset(tmp_nz_val, 0, ell_size * sizeof(T));
    cublasSetVector(ell_size, sizeof(int), tmp_indices, 1, md_spm.indices, 1);
    cublasSetVector(ell_size, sizeof(T), tmp_nz_val, 1, md_spm.nz_val, 1);
    free(tmp_indices);
    free(tmp_nz_val);
    #endif
    
    for(int i = 0; i < md_spm.ncol_per_row; i++)
    {
        #if USE_ALIGNED_OPT_SPM
        cublasSetVector(n_per_thr, sizeof(int), h_spm->indices + i * h_spm->stride + thread_start, 1, md_spm.indices + i * md_spm.stride, 1);
        cublasSetVector(n_per_thr, sizeof(T), h_spm->nz_val + i * h_spm->stride + thread_start, 1, md_spm.nz_val + i * md_spm.stride, 1);    
        #else
        cublasSetVector(n_per_thr, sizeof(int), h_spm->indices + i * h_spm->nrow + thread_start, 1, md_spm.indices + i * n_per_thr, 1);
        cublasSetVector(n_per_thr, sizeof(T), h_spm->nz_val + i * h_spm->nrow + thread_start, 1, md_spm.nz_val + i * n_per_thr, 1);    
        #endif
    }
    
    return md_spm;
}

template <class T>
void md_dealloc_matrix(ell_format<T> *md_spm)
{
    cublasFree(md_spm->indices);
    cublasFree(md_spm->nz_val);
}

template <class T>
void spmv_ell_cpu(int *h_ind, T *h_val, T *h_vec, T *h_result_vec, const int n, const int m, const int ncol_per_row, const int stride)
{
    for(int i = 0; i < m; i++) 
    {   
        T sum = 0.0;
        for(int j = 0; j < ncol_per_row; j++) 
        {
            int index = stride * j + i; 
            T val = h_val[index];
            if (val != 0.0)
                sum += val * h_vec[h_ind[index]];
        }
        h_result_vec[i] = sum;
    }    
}

template <class T>
void spmv_cpu(ell_format<T> *h_spm, T *h_vec, T *h_result_vec)
{
    int n = h_spm->ncol;
    int m = h_spm->nrow;
    int ncol_per_row = h_spm->ncol_per_row;  
    #if USE_ALIGNED_OPT_SPM
    int stride = h_spm->stride;
    #else
    int stride = h_spm->nrow;
    #endif
    spmv_ell_cpu<T>(h_spm->indices, h_spm->nz_val, h_vec, h_result_vec, n, m, ncol_per_row, stride);
}

template <class T>
void spmv_single_gpu(ell_format<T> *d_spm, T *d_result_vec, T *d_vec)
{
    int m = d_spm->nrow;
    int ncol_per_row = d_spm->ncol_per_row;
    #if USE_ALIGNED_OPT_SPM
    int stride = d_spm->stride;
    #else
    int stride = d_spm->nrow;
    #endif
    unsigned int dimBlock = DIMBLOCK;
    const dim3 dimGrid = large_grid(m, dimBlock);
	
    #if USE_TEXTURE
    bind_vec(d_vec);
    #endif    
    spmv_ell_gpu<T><<<dimGrid, dimBlock>>>(d_spm->indices, d_spm->nz_val, d_result_vec, d_vec, m, ncol_per_row, stride);
    #if USE_TEXTURE
    unbind_vec(d_vec);
    #endif     
}

template <class T>
void spmv_multi_gpu(ell_format<T> *md_spm, T *d_result_vec, T *d_vec, const int n_per_thr)
{
    int ncol_per_row = md_spm->ncol_per_row;
    #if USE_ALIGNED_OPT_SPM
    int stride = md_spm->stride;
    #else
    int stride = n_per_thr;
    #endif
    unsigned int dimBlock = DIMBLOCK;
    const dim3 dimGrid = large_grid(n_per_thr, dimBlock);

    #if USE_TEXTURE
    bind_vec(d_vec);
    #endif     
    spmv_ell_gpu<T><<<dimGrid, dimBlock>>>(md_spm->indices, md_spm->nz_val, d_result_vec, d_vec, n_per_thr, ncol_per_row, stride);
    #if USE_TEXTURE
    unbind_vec(d_vec);
    #endif     
}

template <class T>
void preconditioner_Jacobi_ell_cpu(int *h_ind, T *h_val, T *r, T *h, const int n, const int ncol_per_row, const int stride)
{
    for(int i = 0; i < n; i++)
    {   
        T r_val = r[i];
        for(int j = 0; j< ncol_per_row; j++) 
        {
            int index = stride * j + i; 
            if(i == h_ind[index]) 
            {   
                T val_temp = h_val[index];
                T val = (val_temp == 0.0) ? 1.0 : 1.0 / val_temp;
                r_val = r_val * val;
            }
        }
        h[i] =  r_val;
    }    
}

template <class T>
void precon_ssor_Lower_ell_cpu(int *h_ind, T *h_val, T *x, T *y, const int n, const int ncol_per_row, const int stride, const T omega)
{
    for(int i = 0; i < n; i++)
    {
        T sum = 0.0;
        T diag;
        for(int j = 0; j < ncol_per_row; j++) 
        {
            int index = stride * j + i; 
            int col_num = h_ind[index];
            T val = h_val[index];
            if(col_num < i)
                sum -= val * y[col_num];
            if(col_num == i)
                diag = val;
        }
        y[i] = (x[i] - sum) * omega / diag;
    }    
}

template <class T>
void precon_ssor_Diag_ell_cpu(int *h_ind, T *h_val, T *x, const int n, const int ncol_per_row, const int stride, const T omega)
{
    for(int i = 0; i < n; i++)
    {   
        for(int j = 0; j< ncol_per_row; j++) 
        {
            int index = stride * j + i; 
            if(i == h_ind[index]) 
                x[i] *= h_val[index] / omega;
        }
    }    
}

template <class T>
void precon_ssor_Upper_ell_cpu(int *h_ind, T *h_val, T *x, T *y, const int n, const int ncol_per_row, const int stride, const T omega)
{
    for(int i = n - 1; i >= 0; i--)
    {
        T sum = 0.0;
        T diag;
        for(int j = 0; j < ncol_per_row; j++) 
        {
            int index = stride * j + i; 
            int col_num = h_ind[index];
            T val = h_val[index];
            if(col_num > i)
                sum -= val * y[col_num];
            if(col_num == i)
                diag = val;
        }
        y[i] = (x[i] - sum) * omega / diag;        
    }
}

template <class T>
void preconditioner_ssor_ell_cpu(int *h_ind, T *h_val, T *r, T *h, const int n, const int ncol_per_row, const int stride)
{
    T omega = OMEGA;
    T *temp = (T *)malloc(sizeof(T) * n);
    precon_ssor_Lower_ell_cpu<T>(h_ind, h_val, r, temp, n, ncol_per_row, stride, omega);
    precon_ssor_Diag_ell_cpu<T>(h_ind, h_val, temp, n, ncol_per_row, stride, omega);
    precon_ssor_Upper_ell_cpu<T>(h_ind, h_val, temp, h, n, ncol_per_row, stride, omega);
    sca_mul_vec<T>(n, 2.0 - omega, h, h);
    free(temp);
}

template <class T>
void preconditioner_cpu(char *precon_name, ell_format<T> *h_spm, T *r, T *h)
{
    int n = h_spm->ncol;
    int ncol_per_row = h_spm->ncol_per_row; 
    #if USE_ALIGNED_OPT_SPM
    int stride = h_spm->stride;
    #else
    int stride = h_spm->nrow;
    #endif
   
    #if USE_FLOAT_PRECON
    int ell_size = stride * ncol_per_row; 
    float *h_val = (float *)malloc(ell_size * sizeof(float));
    float *r_float = (float *)malloc(n * sizeof(float));
    float *h_float = (float *)malloc(n * sizeof(float));
    std::transform(h_spm->nz_val, h_spm->nz_val + ell_size, h_val, convert<T, float>());
    std::transform(r, r + n, r_float, convert<T, float>());
    std::transform(h, h + n, h_float, convert<T, float>());
    if(!strcmp(precon_name, "Jacobi"))
    preconditioner_Jacobi_ell_cpu<float>(h_spm->indices, h_val, r_float, h_float, n, ncol_per_row, stride);
    else if(!strcmp(precon_name, "SSOR"))
    preconditioner_ssor_ell_cpu<float>(h_spm->indices, h_val, r_float, h_float, n, ncol_per_row, stride);
    std::transform(h_float, h_float + n, h, convert<float, T>());
    free(h_val);
    free(r_float);
    free(h_float);
    #else
    if(!strcmp(precon_name, "Jacobi"))
    preconditioner_Jacobi_ell_cpu<T>(h_spm->indices, h_spm->nz_val, r, h, n, ncol_per_row, stride);
    else if(!strcmp(precon_name, "SSOR"))
    preconditioner_ssor_ell_cpu<T>(h_spm->indices, h_spm->nz_val, r, h, n, ncol_per_row, stride);
    #endif
}

template <class T>
void preconditioner_single_gpu(char *precon_name, ell_format<T> *d_spm, T *r, T *h)
{
    int n = d_spm->ncol;
    int ncol_per_row = d_spm->ncol_per_row; 
    #if USE_ALIGNED_OPT_SPM
    int stride = d_spm->stride;
    #else
    int stride = d_spm->nrow;
    #endif
    unsigned int dimBlock = DIMBLOCK;
    const dim3 dimGrid = large_grid(n, dimBlock);

    #if USE_FLOAT_PRECON
    int ell_size = stride * ncol_per_row; 
    float *d_val = 0;
    float *r_float = 0;
    float *h_float = 0;
    CUDA_SAFE_CALL(cudaMalloc((void **)&d_val, ell_size * sizeof(float)));
    CUDA_SAFE_CALL(cudaMalloc((void **)&r_float, n * sizeof(float)));
    CUDA_SAFE_CALL(cudaMalloc((void **)&h_float, n * sizeof(float)));
    const dim3 dimGrid_val = large_grid(ell_size, dimBlock);
    transform_gpu<T, float><<<dimGrid_val, dimBlock>>>(d_spm->nz_val, d_val, ell_size);
    transform_gpu<T, float><<<dimGrid, dimBlock>>>(r, r_float, n);
    transform_gpu<T, float><<<dimGrid, dimBlock>>>(h, h_float, n);
    #if USE_TEXTURE
    bind_vec(r_float);
    #endif  
    if(!strcmp(precon_name, "Jacobi"))   
    preconditioner_ell_single_gpu<float><<<dimGrid, dimBlock>>>(d_spm->indices, d_val, r_float, h_float, n, ncol_per_row, stride);
    else if(!strcmp(precon_name, "SSOR"))
    preconditioner_ell_single_gpu<float><<<dimGrid, dimBlock>>>(d_spm->indices, d_val, r_float, h_float, n, ncol_per_row, stride);
    #if USE_TEXTURE
    unbind_vec(r_float);
    #endif     
    transform_gpu<float, T><<<dimGrid, dimBlock>>>(h_float, h, n);
    cublasFree(d_val);
    cublasFree(r_float);
    cublasFree(h_float);
    #else
    #if USE_TEXTURE
    bind_vec(r);
    #endif    
    if(!strcmp(precon_name, "Jacobi"))  
    preconditioner_ell_single_gpu<T><<<dimGrid, dimBlock>>>(d_spm->indices, d_spm->nz_val, r, h, n, ncol_per_row, stride);
    else if(!strcmp(precon_name, "SSOR"))
    preconditioner_ell_single_gpu<T><<<dimGrid, dimBlock>>>(d_spm->indices, d_spm->nz_val, r, h, n, ncol_per_row, stride);    
    #if USE_TEXTURE
    unbind_vec(r);
    #endif    
    #endif    
}

template <class T>
void preconditioner_multi_gpu(char *precon_name, ell_format<T> *md_spm, T *r, T *h, const int n_per_thr, const int nnzero_per_thr, const int thread_start)
{
    int ncol_per_row = md_spm->ncol_per_row; 
    #if USE_ALIGNED_OPT_SPM
    int stride = md_spm->stride;
    #else
    int stride = n_per_thr;
    #endif
    unsigned int dimBlock = DIMBLOCK;
    const dim3 dimGrid = large_grid(n_per_thr, dimBlock);

    #if USE_FLOAT_PRECON
    float *md_val = 0;
    float *r_float = 0;
    float *h_float = 0;
    CUDA_SAFE_CALL(cudaMalloc((void **)&md_val, nnzero_per_thr * sizeof(float)));
    CUDA_SAFE_CALL(cudaMalloc((void **)&r_float, n_per_thr * sizeof(float)));
    CUDA_SAFE_CALL(cudaMalloc((void **)&h_float, n_per_thr * sizeof(float)));
    const dim3 dimGrid_val = large_grid(nnzero_per_thr, dimBlock);
    transform_gpu<T, float><<<dimGrid_val, dimBlock>>>(md_spm->nz_val, md_val, nnzero_per_thr);
    transform_gpu<T, float><<<dimGrid, dimBlock>>>(r, r_float, n_per_thr);
    transform_gpu<T, float><<<dimGrid, dimBlock>>>(h, h_float, n_per_thr);
    #if USE_TEXTURE
    bind_vec(r_float);
    #endif  
    if(!strcmp(precon_name, "Jacobi"))    
    preconditioner_ell_multi_gpu<float><<<dimGrid, dimBlock>>>(md_spm->indices, md_val, r_float, h_float, n_per_thr, ncol_per_row, stride, thread_start);
    else if(!strcmp(precon_name, "SSOR"))
    preconditioner_ell_multi_gpu<float><<<dimGrid, dimBlock>>>(md_spm->indices, md_val, r_float, h_float, n_per_thr, ncol_per_row, stride, thread_start);
    #if USE_TEXTURE
    unbind_vec(r_float);
    #endif     
    transform_gpu<float, T><<<dimGrid, dimBlock>>>(h_float, h, n_per_thr);
    cublasFree(md_val);
    cublasFree(r_float);
    cublasFree(h_float);
    #else
    #if USE_TEXTURE
    bind_vec(r);
    #endif     
    if(!strcmp(precon_name, "Jacobi"))  
    preconditioner_ell_multi_gpu<T><<<dimGrid, dimBlock>>>(md_spm->indices, md_spm->nz_val, r, h, n_per_thr, ncol_per_row, stride, thread_start);
    else if(!strcmp(precon_name, "SSOR"))
    preconditioner_ell_multi_gpu<T><<<dimGrid, dimBlock>>>(md_spm->indices, md_spm->nz_val, r, h, n_per_thr, ncol_per_row, stride, thread_start);
    #if USE_TEXTURE
    unbind_vec(r);
    #endif     
    #endif    
}

#endif /* __ELL_FORMAT_H__ */
