/*
 * CUDA for Sparse Solver
 * Author: Luo Li     li.luo@sub.siat.ac.cn
 */ 

#define BLOCKDIM 512
/* If the number of blocks is too large, they are divided into 2D subBlocks, this line returns the innel index that 
   deals with both 1D or 2D cases in kernel function. The corresponding division of blockDim and gridDim for input
   <<< >>> must be done outside. */
#define large_grid_thread_id() (__umul24(blockDim.x, blockIdx.x + __umul24(blockIdx.y, gridDim.x)) + threadIdx.x)

template<class T>
__global__ void spmv_csr_single_gpu(int *d_row, int *d_col, T *d_val, T *d_result_vec, T *d_vec, const int interval_size, const int n)
{
	const int threads_per_row = interval_size;
	const int thread_id = large_grid_thread_id();
	const int row_in_all = thread_id / threads_per_row;//row number of each thread in grid
	const int thread_id_in_row = threadIdx.x % threads_per_row;

	//extern __shared__ T val[];
    extern __shared__ char __val[];
    T *const val = reinterpret_cast<T *>(__val);
    
	val[threadIdx.x] = 0.0;
	if(row_in_all < n)
    {
        const int start = d_row[row_in_all];
        const int end = d_row[row_in_all + 1];
        for(int k = start + thread_id_in_row; k < end ; k += threads_per_row)
        {
            #if USE_TEXTURE
            val[threadIdx.x] += fetch_vec(d_col[k], d_vec) * d_val[k];
            #else
            val[threadIdx.x] += d_vec[d_col[k]] * d_val[k];
            #endif
        }
        //fold form
        /*for(int i = threads_per_row / 2; i >= 1; i /= 2)
        {
            if(thread_id_in_row < i)
            val[threadIdx.x] += val[threadIdx.x + i];
            __syncthreads();
        }
        if(thread_id_in_row == 0) 
        d_result_vec[row_in_all] = val[threadIdx.x];
        */
        //unfold form
        /*
        if(thread_id_in_row < 16)
            val[threadIdx.x] += val[threadIdx.x + 16];
            __syncthreads();*/
        if(thread_id_in_row < 8)
            val[threadIdx.x] += val[threadIdx.x + 8];
            __syncthreads();
        if(thread_id_in_row < 4)
            val[threadIdx.x] += val[threadIdx.x + 4];
            __syncthreads();
        if(thread_id_in_row < 2)
            val[threadIdx.x] += val[threadIdx.x + 2];
            __syncthreads();
        if(thread_id_in_row < 1)
        {
            val[threadIdx.x] += val[threadIdx.x + 1];
            __syncthreads();
            d_result_vec[row_in_all] = val[threadIdx.x];
        }
	}
}

template<class T>
__global__ void spmv_csr_multi_gpu(int *d_row, int *d_col, T *d_val, T *d_result_vec, T *d_vec, const int interval_size, const int n)
{
    const int threads_per_row = interval_size;
    const int thread_id = large_grid_thread_id();
    const int row_in_all = thread_id / threads_per_row;//row number of each thread in grid
    const int thread_id_in_row = threadIdx.x % threads_per_row;	

    //extern __shared__ T val[];
    extern __shared__ char __val[];
    T *const val = reinterpret_cast<T *>(__val);
	
	val[threadIdx.x] = 0.0;
	if(row_in_all < n)
    {
        const int offset = d_row[0];
        const int start = d_row[row_in_all] - offset;
        const int end = d_row[row_in_all + 1] - offset;        
        for(int k = start + thread_id_in_row; k < end ; k += threads_per_row)
        {	
            #if USE_TEXTURE
            val[threadIdx.x] += fetch_vec(d_col[k], d_vec) * d_val[k];
            #else
            val[threadIdx.x] += d_vec[d_col[k]] * d_val[k];
            #endif
        }
        //fold form
        /*
        for(int i = threads_per_row / 2; i >= 1; i /= 2)
        {
            if(thread_id_in_row < i)
            val[threadIdx.x] += val[threadIdx.x + i];
            __syncthreads();
        }
        if(thread_id_in_row == 0) 
        d_result_vec[row_in_all] = val[threadIdx.x];
        */
        //unfold form
        /*
        if(thread_id_in_row < 16)
            val[threadIdx.x] += val[threadIdx.x + 16];
            __syncthreads();*/
        if(thread_id_in_row < 8)
            val[threadIdx.x] += val[threadIdx.x + 8];
            __syncthreads();
        if(thread_id_in_row < 4)
            val[threadIdx.x] += val[threadIdx.x + 4];
            __syncthreads();
        if(thread_id_in_row < 2)
            val[threadIdx.x] += val[threadIdx.x + 2];
            __syncthreads();
        if(thread_id_in_row < 1)
        {
            val[threadIdx.x] += val[threadIdx.x + 1];
            __syncthreads();
            d_result_vec[row_in_all] = val[threadIdx.x];
        }
	}
}

////////////////Unuse
/**/
template<class T>
__global__ void preconditioner_Jacobi_vec_csr_single_gpu(int *d_row, int *d_col, T *d_val, T *r, T *h, const int interval_size, const int n)
{
    const int threads_per_row = interval_size;
    const int thread_id = large_grid_thread_id();
    const int row_in_all = thread_id / threads_per_row;//row number of each thread in grid
    const int thread_id_in_row = threadIdx.x % threads_per_row;    
    
    if(row_in_all < n)
    {
        #if USE_TEXTURE
        h[row_in_all] = fetch_vec(row_in_all, r);//若用寄存器则可能不够用
        #else
        h[row_in_all] = r[row_in_all];
        #endif
        const int start = d_row[row_in_all];
        const int end = d_row[row_in_all + 1];
        for(int k = start + thread_id_in_row; k < end; k += threads_per_row)
            if(row_in_all == d_col[k]) 
            {   
                T val_temp = d_val[k];
                T val = (val_temp == 0.0) ? 1.0 : 1.0 / val_temp;
                #if USE_TEXTURE
                h[row_in_all] = fetch_vec(row_in_all, r) * val;            
                #else
                h[row_in_all] = r[row_in_all] * val;
                #endif
            }
    }
}

template<class T>
__global__ void preconditioner_Jacobi_vec_csr_multi_gpu(int *d_row, int *d_col, T *d_val, T *r, T *h, const int interval_size, const int n, const int thread_start)
{
    const int threads_per_row = interval_size;
    const int thread_id = large_grid_thread_id();
    const int row_in_all = thread_id / threads_per_row;//row number of each thread in grid
    const int thread_id_in_row = threadIdx.x % threads_per_row;    
    const int row_in_mat = thread_start + row_in_all;

    if(row_in_all < n)
    {
        #if USE_TEXTURE
        h[row_in_all] = fetch_vec(row_in_all, r);
        #else
        h[row_in_all] = r[row_in_all];
        #endif        
        const int offset = d_row[0];
        const int start = d_row[row_in_all] - offset;
        const int end = d_row[row_in_all + 1] - offset;
        for(int k = start + thread_id_in_row; k < end; k += threads_per_row)
            if(row_in_mat == d_col[k]) 
            {            
                T val_temp = d_val[k];
                T val = (val_temp == 0.0) ? 1.0 : 1.0 / val_temp;                
                #if USE_TEXTURE
                h[row_in_all] = fetch_vec(row_in_all, r) * val;            
                #else
                h[row_in_all] = r[row_in_all] * val;
                #endif                
            }
    }
}

template<class T>
__global__ void preconditioner_Jacobi_scal_csr_single_gpu(int *d_row, int *d_col, T *d_val, T *r, T *h, const int n)
{
	const int row = large_grid_thread_id();

    if(row < n)
    {
        #if USE_TEXTURE
        T r_val = fetch_vec(row, r);
        #else
        T r_val = r[row];
        #endif
        const int start = d_row[row];
        const int end = d_row[row + 1];
        for(int k = start; k < end; k++)
            if(row == d_col[k]) 
            {   
                T val_temp = d_val[k];
                T val = (val_temp == 0.0) ? 1.0 : 1.0 / val_temp;
                r_val = r_val * val;
            }
       h[row] = r_val;
    }
}

template<class T>
__global__ void preconditioner_Jacobi_scal_csr_multi_gpu(int *d_row, int *d_col, T *d_val, T *r, T *h, const int n, const int thread_start)
{
    const int row = large_grid_thread_id();
    const int row_in_mat = thread_start + row;

    if(row < n)
    {
        #if USE_TEXTURE
        T r_val = fetch_vec(row, r);
        #else
        T r_val = r[row];
        #endif        
        const int offset = d_row[0];
        const int start = d_row[row] - offset;
        const int end = d_row[row + 1] - offset;
        for(int k = start; k < end; k++)
            if(row_in_mat == d_col[k]) 
            {   
                T val_temp = d_val[k];
                T val = (val_temp == 0.0) ? 1.0 : 1.0 / val_temp;
                r_val = r_val * val;
            }
        h[row] = r_val;
    }
}



template<class T>
__global__ void precon_ssor_Diag_csr_single_gpu(int *d_row, int *d_col, T *d_val, T *x, const int n, const T omega)
{
    const int row = large_grid_thread_id();
    if(row < n)
    {
        const int start = d_row[row];
        const int end = d_row[row + 1];
        for(int k = start; k < end; k++)
            if(row == d_col[k]) 
                x[row] *= d_val[k] / omega;
    }    
}

////////////////////////////////////////////

template<class T>
__global__ void spmv_ell_gpu(int *d_ind, T *d_val, T *d_result_vec, T *d_vec, const int m, const int ncol_per_row, const int stride)
{
	const int row = large_grid_thread_id();
    if(row < m) 
    {
        T sum = 0.0;
        for(int j = 0; j < ncol_per_row; j++) 
        {
            const int index = d_ind[stride * j + row];
            T val = d_val[stride * j + row];
            if(val != 0.0) 
            {
                #if USE_TEXTURE
                sum += val * fetch_vec(index, d_vec);
                #else
                sum += val * d_vec[index];
                #endif
            }
        }
        d_result_vec[row] = sum;
    }
}

template<class T>
__global__ void preconditioner_ell_single_gpu(int *d_ind, T *d_val, T *r, T *h, const int n, const int ncol_per_row, const int stride)
{
	const int row = large_grid_thread_id();
    if(row < n) 
    {
        #if USE_TEXTURE
        T r_val = fetch_vec(row, r);
        #else
        T r_val = r[row];
        #endif        
        for(int j = 0; j < ncol_per_row; j++) 
        {
            const int index = stride * j + row; 
            if(row == d_ind[index]) 
            {   
                T val_temp = d_val[index];
                T val = (val_temp == 0.0) ? 1.0 : 1.0 / val_temp;
                r_val = r_val * val;
            }            
        }
        h[row] = r_val;
    }
}

template<class T>
__global__ void preconditioner_ell_multi_gpu(int *d_ind, T *d_val, T *r, T *h, const int n, const int ncol_per_row, const int stride, const int thread_start)
{
	const int row = large_grid_thread_id();
    const int row_in_mat = thread_start + row;
    if(row < n) 
    {
        #if USE_TEXTURE
        T r_val = fetch_vec(row, r);
        #else
        T r_val = r[row];
        #endif        
        for(int j = 0; j < ncol_per_row; j++) 
        {
            const int index = stride * j + row; 
            if(row_in_mat == d_ind[index]) 
            {   
                T val_temp = d_val[index];
                T val = (val_temp == 0.0) ? 1.0 : 1.0 / val_temp;
                r_val = r_val * val;
            }            
        }
        h[row] = r_val;
    }
}

////////////////////////////////////////////

template<class T>
__global__ void precon_Jacobi_gpu(T *diag, T *r, T *h, const int n)
{
    const int row = large_grid_thread_id();
    if(row < n) 
    {
        #if USE_TEXTURE
        T r_val = fetch_vec(row, r);
        #else
        T r_val = r[row];
        #endif        
        h[row] = r_val * diag[row];
    }
} 
/////////////////////Preconditioner SSOR, can be implemented completely with Forward/Backward Elimination.
/* 	SSOR1: Column-major approach. 
	For Lower triangle. Forward Elimination.
	For single GPU(currently no multi-GPU version for SSOR1 supported)
*/
template<class T>
__global__ void precon_ssor_lower_single_gpu(int *col_ptr, int *row_ind, T *nz_val, T *diag, T *x, const int n, const T omega)
{
    const int thread_id = large_grid_thread_id();
	/* Traversing columns */
    for(int k = 0; k < n - 1; k++)
    {
        if(thread_id == 0)
		/* The read-only array diag can be binded to Texture, but x can't, since it would be changed. */
            x[k] *= omega / diag[k];
		/* Synchronize x to guarantee that all threads access below has the most current data x_tmp. */	
        __syncthreads(); 

        T x_tmp = x[k];
        int start = col_ptr[k];
		/* Number of nnzeros of this column */
        int num = col_ptr[k + 1] - start;            
        if(thread_id < num)   
        {
			/* Current index for nnzero val_array */
            int jj = start + thread_id;
			/* row_ind[jj] stands for row index of this nnzero val.
			Both row_ind[jj] and nz_val[jj] are coalesced access, but x is not. 
			x is stored in global memory, which can't be accelerated by Texture, Register or Share memory.
			Unuse Register: Since we don't know which row is nnzero, some element is at this column but is not at next column. 
			Unuse Share memory: Since we need to judge if row_ind[jj] is at the current local section, also, 
			we need to map the global index into local index. And, the column maybe very sparse that this way would 
			waste too many share memory. */
                x[row_ind[jj]] += x_tmp * nz_val[jj];
        }
		/* Synchronize when finish this column */
        __syncthreads();
    }
	/* The last element of diag, no modify needed */
    if(thread_id == 0)
        x[n - 1] *= omega / diag[n - 1]; 
}

/* 	SSOR1: Column-major approach. 
	For Diag.
	For single GPU(currently no multi-GPU version for SSOR1 supported)
*/
template<class T>
__global__ void precon_ssor_diag_single_gpu(T *diag, T *x, const int n, const T omega)
{
    const int row = large_grid_thread_id();
    if(row < n) 
        x[row] *= diag[row] / omega;
}

/* 	SSOR1: Column-major approach. 
	For Upper triangle. Backward Elimination.
	For single GPU(currently no multi-GPU version for SSOR1 supported)
*/
template<class T>
__global__ void precon_ssor_upper_single_gpu(int *col_ptr, int *row_ind, T *nz_val, T *diag, T *x, const int n, const T omega)
{
    const int thread_id = large_grid_thread_id();
	/* Backward traversing columns */
    for(int k = n - 1; k > 0; k--)
    {
        if(thread_id == 0)
            x[k] *= omega / diag[k];
        __syncthreads(); 

        T x_tmp = x[k];
		/* */
        int ii = n - 1 - k;
        int start = col_ptr[ii];
        int num = col_ptr[ii + 1] - start;            
        if(thread_id < num)   
        {
            int jj = start + thread_id;
                x[row_ind[jj]] += x_tmp * nz_val[jj];
        }
        __syncthreads();
    }
    if(thread_id == 0)
        x[0] *= omega / diag[0];
}

template<class T>
__global__ void ssor_subtri_lower(int *col_ptr, int *row_ind, T *nz_val, T *diag, T *x, const int offset, int n_subtri, const int n, const T omega)
{
    const int col = offset + threadIdx.x;

    __shared__ T seg[BLOCKDIM];
    //__shared__ char __seg[BLOCKDIM];
    //T *const seg = reinterpret_cast<T *>(__seg);
    __shared__ T dia[BLOCKDIM];
    //__shared__ char __dia[BLOCKDIM];
    //T *const dia = reinterpret_cast<T *>(__dia);
    
    if(col < n)
    {   
        seg[threadIdx.x] = x[col];
        dia[threadIdx.x] = diag[col];
    }
    __syncthreads();

    for(int k = 0; k < n_subtri - 1; k++)
    {
        if(threadIdx.x == 0)
            seg[k] *= omega / dia[k];
        __syncthreads(); 
    
        int start = col_ptr[k];
        int num = col_ptr[k + 1] - start;            
        if(threadIdx.x < num)   
        {
            int jj = start + threadIdx.x;
            int row = row_ind[jj];
            //#if USE_TEXTURE
            //seg[row] += seg[k] * fetch_vec(jj, nz_val);
            //#else
            seg[row] += seg[k] * nz_val[jj];
            //#endif            
        }
        __syncthreads();
    }
    if(threadIdx.x == 0)
        seg[n_subtri - 1] *= omega / dia[n_subtri - 1];
    
    if(col < n)
        x[col] = seg[threadIdx.x];
    __syncthreads();
}

template<class T>
__global__ void ssor_diag(T *diag, T *x, T *y, const int offset, const int n, const T omega)
{
    const int col = offset + threadIdx.x;

    #if 0
    __shared__ T dia[BLOCKDIM];
    //__shared__ char __dia[BLOCKDIM];
    //T *const dia = reinterpret_cast<T *>(__dia);
    if(col < n)
    {   
        #if USE_TEXTURE
        dia[threadIdx.x] = fetch_vec(col, diag);
        #else
        dia[threadIdx.x] = diag[col];
        #endif
    }
    __syncthreads();
    if(col < n)
        y[col] = x[col] * dia[threadIdx.x] / omega;
    #else
    if(col < n)
    {
        #if USE_TEXTURE
        y[col] = x[col] * fetch_vec(col, diag) / omega;         
        #else
        y[col] = x[col] * diag[col] / omega; 
        #endif
    }
    #endif
}

template<class T>
__global__ void ssor_modify_lower(int *d_row, int *d_col, T *d_val, T *x, const int offset_subtri, const int offset_modify, const int dimBlock, const int interval_size, const int n)
{
    const int threads_per_row = interval_size;
    const int thread_id = large_grid_thread_id();
    const int row_in_mat = thread_id / threads_per_row + offset_modify;//row number of each thread in matrix
    const int thread_id_in_row = threadIdx.x % threads_per_row;
    const int row_subtri = offset_subtri + threadIdx.x;
   
    __shared__ T seg[BLOCKDIM];
    //__shared__ char __seg[BLOCKDIM];
    //T *const seg = reinterpret_cast<T *>(__seg);
    __shared__ T x_modify[BLOCKDIM];
    //__shared__ char __x_modify[BLOCKDIM];
    //T *const x_modify = reinterpret_cast<T *>(__x_modify);

    if(row_subtri < n)
        x_modify[threadIdx.x] = x[row_subtri];
    __syncthreads();
    seg[threadIdx.x] = 0.0;

    if(row_in_mat < n)
    {
        const int start = d_row[row_in_mat];
        const int end = d_row[row_in_mat + 1];
        for(int k = start + thread_id_in_row; k < end ; k += threads_per_row)
        {
            int col = d_col[k];
            if(col >= offset_subtri && col < offset_modify)
            {
                int col_in_block = col % dimBlock;
                seg[threadIdx.x] += x_modify[col_in_block] * d_val[k];
            }
        }
        //fold form
        /*
        for(int i = threads_per_row / 2; i >= 1; i /= 2)
        {
            if(thread_id_in_row < i)
            seg[threadIdx.x] += seg[threadIdx.x + i];
            __syncthreads();
        }
        if(thread_id_in_row == 0) 
            x[row_in_mat] += seg[threadIdx.x];
        */
        //unfold form
        /*
        if(thread_id_in_row < 16)
            seg[threadIdx.x] += seg[threadIdx.x + 16];
            __syncthreads();*/
        if(thread_id_in_row < 8)
            seg[threadIdx.x] += seg[threadIdx.x + 8];
            __syncthreads();
        if(thread_id_in_row < 4)
            seg[threadIdx.x] += seg[threadIdx.x + 4];
            __syncthreads();
        if(thread_id_in_row < 2)
            seg[threadIdx.x] += seg[threadIdx.x + 2];
            __syncthreads();
        if(thread_id_in_row < 1)
        {
            seg[threadIdx.x] += seg[threadIdx.x + 1];
            __syncthreads();
            x[row_in_mat] += seg[threadIdx.x];
        }
    }
}

template<class T>
__global__ void ssor_subtri_upper(int *col_ptr, int *row_ind, T *nz_val, T *diag, T *x, const int offset, int n_subtri, const int n, const T omega)
{
    const int col_re = offset + threadIdx.x;
    const int col = n - 1 - col_re; 

    __shared__ T seg[BLOCKDIM];
    //__shared__ char __seg[BLOCKDIM];
    //T *const seg = reinterpret_cast<T *>(__seg);
    __shared__ T dia[BLOCKDIM];
    //__shared__ char __dia[BLOCKDIM];
    //T *const dia = reinterpret_cast<T *>(__dia);
    
    if(col_re < n)
    {   
        seg[threadIdx.x] = x[col];
        dia[threadIdx.x] = diag[col];
    }
    __syncthreads();

    for(int k = 0; k < n_subtri - 1; k++)
    {
        if(threadIdx.x == 0)
            seg[k] *= omega / dia[k];
        __syncthreads(); 
    
        int start = col_ptr[k];
        int num = col_ptr[k + 1] - start;            
        if(threadIdx.x < num)   
        {
            int jj = start + threadIdx.x;
            int row = row_ind[jj];
            //#if USE_TEXTURE
            //seg[row] += seg[k] * fetch_vec(jj, nz_val);
            //#else
            seg[row] += seg[k] * nz_val[jj];
            //#endif            
        }
        __syncthreads();
    }
    if(threadIdx.x == 0)
        seg[n_subtri - 1] *= omega / dia[n_subtri - 1];
    
    if(col_re < n)
        x[col] = seg[threadIdx.x];
    __syncthreads();
}

template<class T>
__global__ void ssor_modify_upper(int *d_row, int *d_col, T *d_val, T *x, const int offset_subtri, const int offset_modify, const int dimBlock, const int interval_size, const int n)
{
    const int threads_per_row = interval_size;
    const int thread_id = large_grid_thread_id();
    const int row_in_mat_re = thread_id / threads_per_row + offset_modify;//row number of each thread in matrix
    const int row_in_mat = n - 1 - row_in_mat_re;
    const int thread_id_in_row = threadIdx.x % threads_per_row;
    const int row_subtri = offset_subtri + threadIdx.x;
    
    __shared__ T seg[BLOCKDIM];
    //__shared__ char __seg[BLOCKDIM];
    //T *const seg = reinterpret_cast<T *>(__seg);
    __shared__ T x_modify[BLOCKDIM];
    //__shared__ char __x_modify[BLOCKDIM];
    //T *const x_modify = reinterpret_cast<T *>(__x_modify);

    if(row_subtri < n)
    {
        const int row_modify = n - 1 - row_subtri;
        x_modify[threadIdx.x] = x[row_modify];
    }
    __syncthreads();
    seg[threadIdx.x] = 0.0;

    if(row_in_mat_re < n)
    {
        const int start = d_row[row_in_mat];
        const int end = d_row[row_in_mat + 1];
        for(int k = start + thread_id_in_row; k < end ; k += threads_per_row)
        {
            int col = d_col[k];
            int col_re = n - 1 - col;
            if(col_re > offset_subtri && col_re <= offset_modify)
            {
                int col_re_in_block = col_re % dimBlock;
                seg[threadIdx.x] += x_modify[col_re_in_block] * d_val[k];
            }
        }
        //fold form
        /*
        for(int i = threads_per_row / 2; i >= 1; i /= 2)
        {
            if(thread_id_in_row < i)
            seg[threadIdx.x] += seg[threadIdx.x + i];
            __syncthreads();
        }
        if(thread_id_in_row == 0) 
            x[row_in_mat] += seg[threadIdx.x];
        */
        //unfold form
        /*
        if(thread_id_in_row < 16)
            seg[threadIdx.x] += seg[threadIdx.x + 16];
            __syncthreads();*/
        if(thread_id_in_row < 8)
            seg[threadIdx.x] += seg[threadIdx.x + 8];
            __syncthreads();
        if(thread_id_in_row < 4)
            seg[threadIdx.x] += seg[threadIdx.x + 4];
            __syncthreads();
        if(thread_id_in_row < 2)
            seg[threadIdx.x] += seg[threadIdx.x + 2];
            __syncthreads();
        if(thread_id_in_row < 1)
        {
            seg[threadIdx.x] += seg[threadIdx.x + 1];
            __syncthreads();
            x[row_in_mat] += seg[threadIdx.x];
        }
    }
}

////////////////////////////////////////////

template<class T1, class T2>
__global__ void transform_gpu(T1 *a, T2 *b, const int n)
{
	int thread_id = large_grid_thread_id();
    if(thread_id < n)
    {
        b[thread_id] = static_cast<T2>(a[thread_id]);
    }
}
