/*
 * CUDA for Sparse Solver
 * Author: Luo Li     li.luo@sub.siat.ac.cn
 */ 

#ifndef __PCG_DEVICE_H__
#define __PCG_DEVICE_H__

#include "cutils.h"

template <class Matrix_T, class Precon_T, class T>
T pcg_device(FILE *fpDat, Matrix_T *h_spm, Precon_T *h_precon, T *h_b, T *h_x, const int max_iter, const T error_rel, int *iter)
{
    int n;
    if(h_spm->nrow == h_spm->ncol) 
        n = h_spm->nrow;
    else
    {
        printf("PCG_DEVICE: Nonsquare matrix doesn't supported.\n");
        exit(EXIT_FAILURE);
    }

	cublasStatus status;
	int k;
	T alpha, beta, rh, error_norm;
    
    timer spmv_timer;
    begin_timer(&spmv_timer); 
    double spmv_old_time, spmv_elapsed_time, spmv_aver_time = 0.0;
    timer precon_timer;
    begin_timer(&precon_timer); 
    double precon_old_time, precon_elapsed_time, precon_aver_time = 0.0;    
    
	unsigned int timer_compu = 0;
	double t_compu_ms = 0.0;
    start_timer(&timer_compu);	

	/* initialize CUBLAS */
	status = cublasInit();
	if(status != CUBLAS_STATUS_SUCCESS)
	{
		fprintf(stderr, "!!!! CUBLAS initialization error\n");
		exit(EXIT_FAILURE);
	}
	
    T *d_b = 0;
    T *d_x = 0;	
    T *p = 0;
    T *h = 0;
    T *r = 0;
    T *Ap = 0;
    T *Ax = 0;
    
    /* allocate and upload device memory */
    Matrix_T d_spm = d_alloc_matrix<T>(h_spm);
    Precon_T d_precon = d_alloc_precon<T>(h_precon);
    CUDA_SAFE_CALL(cudaMalloc((void **)&d_b, n * sizeof(T)));
	CUDA_SAFE_CALL(cudaMalloc((void **)&d_x, n * sizeof(T)));	
    CUDA_SAFE_CALL(cudaMemcpy(d_b, h_b, n * sizeof(T), cudaMemcpyHostToDevice));
    CUDA_SAFE_CALL(cudaMemcpy(d_x, h_x, n * sizeof(T), cudaMemcpyHostToDevice));

	/* allocate temporary vectors */
	cublasAlloc(n, sizeof(T), (void **)&p);
	cublasAlloc(n, sizeof(T), (void **)&h);
	cublasAlloc(n, sizeof(T), (void **)&r);
	cublasAlloc(n, sizeof(T), (void **)&Ap);
	cublasAlloc(n, sizeof(T), (void **)&Ax);

	/* r := b - Ax */
    spmv_single_gpu<T>(&d_spm, r, d_x);
	cublasXaxpy(n, -1.0, d_b, 1, r, 1);
    cublasXscal(n, -1.0, r, 1);
    /* Mh := r */
    //preconditioner_single_gpu<T>("Jacobi", &d_spm, r, h);
    precon_single_gpu<Matrix_T, T>(&d_spm, &d_precon, h_precon, r, h);
	/* p := h */
	cublasXcopy(n, h, 1, p, 1);
    rh = cublasXdot(n, r, 1, h, 1);

    T b_square = cublasXnrm2(n, d_b, 1);
    #if USE_SAME_ERROR
    #if USE_ERROR_REL
    T error_norm_rel = error_rel * error_rel * b_square;
    #endif
    #else
    #if USE_ERROR_REL
    T error_norm_rel = error_rel * error_rel * rh;
    #endif
    #endif

    #if SET_SUPER_MAX_ITER
    for(k = 0; k < SUPER_MAX_ITER; ++k)
    #else
    for(k = 0; k < n; ++k)
    #endif
	{
		/* alpha := r'h / p'Ap */
        spmv_old_time = end_timer(&spmv_timer); 
        spmv_single_gpu<T>(&d_spm, Ap, p);
        spmv_elapsed_time = elapsed_timer(&spmv_timer, spmv_old_time);
        spmv_aver_time += spmv_elapsed_time;        
		alpha = rh / cublasXdot(n, p, 1, Ap, 1);
		/* x := x + alpha p */
		cublasXaxpy(n, alpha, p, 1, d_x, 1);
		/* r := r - alpha Ap */
		cublasXaxpy(n, -alpha, Ap, 1, r, 1);
        /* Mh := r */
        precon_old_time = end_timer(&precon_timer); 
        //preconditioner_single_gpu<T>("Jacobi", &d_spm, r, h);
        precon_single_gpu<Matrix_T, T>(&d_spm, &d_precon, h_precon, r, h);
        precon_elapsed_time = elapsed_timer(&precon_timer, precon_old_time);
        precon_aver_time += precon_elapsed_time;           
		/* beta := r'h_new / r'h_old */
        beta = 1.0 / rh;
        rh = cublasXdot(n, r, 1, h, 1);
		beta *= rh;
		/* p := h + beta p */
		cublasXscal(n, beta, p, 1);
		cublasXaxpy(n, 1.0, h, 1, p, 1);
        #if USE_SAME_ERROR
        #if USE_ERROR_REL
        if((error_norm = cublasXnrm2(n, r, 1)) < error_norm_rel || k > max_iter)
            break;
        #else
        if((error_norm = cublasXnrm2(n, r, 1)) < ERROR_ABS || k > max_iter)
            break;
        #endif
        if(!(k % SHOW_INTERVAL)) printf("SIGLE DEVICE: interation num: %d, error: %f\n", k, error_norm);
        #else
        #if USE_ERROR_REL
        if(rh < error_norm_rel || k > max_iter)
            break;
        #else
        if(rh < ERROR_ABS || k > max_iter)
            break;
        #endif
        if(!(k % SHOW_INTERVAL)) printf("SIGLE DEVICE: interation num: %d, error: %f\n", k, rh);
        #endif        
	}
	stop_timer(timer_compu, &t_compu_ms);
    *iter = k;
	fprintf(fpDat, "The computation time of single GPU is: %f ms\n", t_compu_ms);
    fprintf(fpDat, "The iteration of single GPU is: %d\n", k);

    spmv_aver_time /= (k + 1);
    fprintf(fpDat, "The SPMV time is: %f s\n", spmv_aver_time);
    double Gflops_spmv = (spmv_aver_time == 0.0) ? 0.0 : 
    ((2.0 * h_spm->nnzero) / spmv_aver_time) / 1e9;
    fprintf(fpDat, "The compute GFLOPs for SPMV: %2.2f\n", Gflops_spmv);
    precon_aver_time /= (k + 1);

    double precon_flop = get_precon_flop<T>(h_precon);
    fprintf(fpDat, "The precon time is: %f s\n", precon_aver_time);
    double Gflops_precon = (precon_aver_time == 0.0) ? 0.0 : 
    (precon_flop / precon_aver_time) / 1e9;
    fprintf(fpDat, "The compute GFLOPs for precon: %2.2f\n", Gflops_precon);    

    #if USE_SAME_ERROR
    double Gflops = (t_compu_ms == 0.0) ? 0.0 : 
    ((2.0 * h_spm->nnzero + 7.0 * n + precon_flop + (k + 1) * (2.0 * h_spm->nnzero + 12.0 * n + precon_flop)) / t_compu_ms) / 1e6;
    fprintf(fpDat, "The compute GFLOPs for PCG: %2.2f\n", Gflops);
    #else
    error_norm = cublasXnrm2(n, r, 1);
    double Gflops = (t_compu_ms == 0.0) ? 0.0 : 
    ((2.0 * h_spm->nnzero + 7.0 * n + precon_flop + (k + 1) * (2.0 * h_spm->nnzero + 10.0 * n + precon_flop)) / t_compu_ms) / 1e6;
    fprintf(fpDat, "The compute GFLOPs for PCG: %2.2f\n", Gflops);
    #endif

	/* transfer solution from device to host */
	status = cublasGetVector(n, sizeof(T), d_x, 1, h_x, 1);
	if(status != CUBLAS_STATUS_SUCCESS)
	{
		fprintf(stderr, "!!!! device access error - 'd_x'\n");
		exit(EXIT_FAILURE);
	}
	
	/* free temporary vectors */
    cublasFree(p);
    cublasFree(h);
    cublasFree(r);
	cublasFree(Ap);
	cublasFree(Ax);
	/* free device memory */
    d_dealloc_matrix<T>(&d_spm);  
    d_dealloc_precon<T>(&d_precon);    
    CUDA_SAFE_CALL(cudaFree(d_b));
    CUDA_SAFE_CALL(cudaFree(d_x));
	
	/* shutdown CUBLAS */
	status = cublasShutdown();
	if(status != CUBLAS_STATUS_SUCCESS)
	{
		fprintf(stderr, "!!!! shutdown error\n");
		exit(EXIT_FAILURE);
	}

    #if USE_ERROR_REL
    return sqrt(error_norm / b_square);
    #else
    return sqrt(error_norm / n);
    #endif
}

#endif /* __PCG_DEVICE_H__ */
