/*
 * CUDA for Sparse Solver
 * Author: Luo Li     li.luo@sub.siat.ac.cn
 */ 

#ifndef __PCG_HOST_H__
#define __PCG_HOST_H__

#include "blas.h"

template <class Matrix_T, class Precon_T, class T>
T pcg_host(FILE *fpDat, Matrix_T *h_spm, Precon_T *h_precon, T *h_b, T *h_x, const int max_iter, const T error_rel, int *iter)
{     
    int n;
    if(h_spm->nrow == h_spm->ncol) 
        n = h_spm->nrow;
    else
    {
        printf("PCG_HOST: Nonsquare matrix doesn't supported.\n");
        exit(EXIT_FAILURE);
    } 

    int k;
    T alpha, beta, error_norm, rh;

    unsigned int timer_compu = 0;
    double t_compu_ms = 0.0;    
    start_timer(&timer_compu);     

    /* allocate temporary vectors */
    T *p = (T *)malloc(n * sizeof(T));
    T *h = (T *)malloc(n * sizeof(T));
    T *r = (T *)malloc(n * sizeof(T));
    T *Ap = (T *)malloc(n * sizeof(T)); 
	
	/* r := b - Ax */
    spmv_cpu<T>(h_spm, h_x, r);	
    Xaxpy(n, -1.0, h_b, 1, r, 1);
    Xscal(n, -1.0, r, 1);
    /* Mh := r */
    //preconditioner_cpu<T>("SSOR", h_spm, r, h);
    precon_cpu<Matrix_T, T>(h_spm, h_precon, r, h);
	/* p := h */
    Xcopy(n, h, 1, p, 1);
    rh = Xdot(n, r, 1, h, 1);

    T b_square = Xnrm2(n, h_b, 1); 
    #if USE_SAME_ERROR
    #if USE_ERROR_REL
    T error_norm_rel = error_rel * error_rel * b_square;
    #endif
    #else
    #if USE_ERROR_REL
    T error_norm_rel = error_rel * error_rel * rh;
    #endif
    #endif
	
    #if SET_SUPER_MAX_ITER
    for(k = 0; k < SUPER_MAX_ITER; ++k)
    #else
    for(k = 0; k < n; ++k)
    #endif	
	{
		/* alpha := r'h / p'Ap */
        spmv_cpu<T>(h_spm, p, Ap);	
		alpha = rh / Xdot(n, p, 1, Ap, 1);
		/* x := x + alpha p */
        Xaxpy(n, alpha, p, 1, h_x, 1);
		/* r := r - alpha Ap */
        Xaxpy(n, -alpha, Ap, 1, r, 1);
        /* Mh := r */
        //preconditioner_cpu<T>("SSOR", h_spm, r, h);
        precon_cpu<Matrix_T, T>(h_spm, h_precon, r, h);
		/* beta := r'h_new / r'h_old */
        beta = 1.0 / rh;
        rh = Xdot(n, r, 1, h, 1);
		beta *= rh;
		/* p := h + beta p */
        Xscal(n, beta, p, 1);
        Xaxpy(n, 1.0, h, 1, p, 1);
        #if USE_SAME_ERROR
        #if USE_ERROR_REL
        if((error_norm = Xnrm2(n, r, 1)) < error_norm_rel || k > max_iter)
            break;
        #else
        if((error_norm = Xnrm2(n, r, 1)) < ERROR_ABS || k > max_iter)
            break;        
        #endif
        if(!(k % SHOW_INTERVAL)) printf("HOST: interation num: %d, error: %f\n", k, error_norm);
        #else
        #if USE_ERROR_REL
        if(rh < error_norm_rel || k > max_iter)
            break;
        #else
        if(rh < ERROR_ABS || k > max_iter)
            break;
        #endif
        if(!(k % SHOW_INTERVAL)) printf("HOST: interation num: %d, error: %f\n", k, rh);
        #endif
	}
    stop_timer(timer_compu, &t_compu_ms);
    *iter = k;
    fprintf(fpDat, "The iteration of CPU is: %d\n", k);

    double precon_flop = get_precon_flop<T>(h_precon);
    #if USE_SAME_ERROR
    double Gflops = (t_compu_ms == 0.0) ? 0.0 : 
    ((2.0 * h_spm->nnzero + 5.0 * n + precon_flop + (k + 1) * (2.0 * h_spm->nnzero + 12.0 * n + precon_flop)) / t_compu_ms) / 1e6;
    fprintf(fpDat, "The compute GFLOPs for PCG: %2.2f\n", Gflops);
    #else
    error_norm = Xnrm2(n, r, 1);
    double Gflops = (t_compu_ms == 0.0) ? 0.0 : 
    ((2.0 * h_spm->nnzero + 5.0 * n + precon_flop + (k + 1) * (2.0 * h_spm->nnzero + 10.0 * n + precon_flop)) / t_compu_ms) / 1e6;
    fprintf(fpDat, "The compute GFLOPs for PCG: %2.2f\n", Gflops);
    #endif    

	free(p);
    free(h);
	free(r);
	free(Ap);

	#if USE_ERROR_REL
	return sqrt(error_norm / b_square);
    #else
    return sqrt(error_norm / n);
    #endif
}

#endif /* __PCG_HOST_H__ */
