/*
 * CUDA for Sparse Solver
 * Author: Luo Li     li.luo@sub.siat.ac.cn
 */ 

#ifndef __PGMRES_DEVICE_H__
#define __PGMRES_DEVICE_H__

#include "cutils.h"

template <class Matrix_T, class Precon_T, class T>
T pgmres_device(FILE *fpDat, Matrix_T *h_spm, Precon_T *h_precon, T *h_b, T *h_x, const int max_iter, const T error_rel, int *iter)
{
    int n;
    if(h_spm->nrow == h_spm->ncol) 
        n = h_spm->nrow;
    else
    {
        printf("PGMRES_DEVICE: Nonsquare matrix doesn't supported.\n");
        exit(EXIT_FAILURE);
    }

	cublasStatus status;
    T beta, h, rd, dd, error_norm;
    int j, io, uij, u0j, k; 
    int m = GMRES_RESTART;
    
	unsigned int timer_compu = 0;
	double t_compu_ms = 0.0;
    start_timer(&timer_compu);	

	/* initialize CUBLAS */
	status = cublasInit();
	if(status != CUBLAS_STATUS_SUCCESS)
	{
		fprintf(stderr, "!!!! CUBLAS initialization error\n");
		exit(EXIT_FAILURE);
	}
	
    T *d_b = 0;
    T *d_x = 0;	
    T *V = 0;
    T *r = 0;
    
    /* allocate and upload device memory */
    Matrix_T d_spm = d_alloc_matrix<T>(h_spm);
    Precon_T d_precon = d_alloc_precon<T>(h_precon);
    CUDA_SAFE_CALL(cudaMalloc((void **)&d_b, n * sizeof(T)));
	CUDA_SAFE_CALL(cudaMalloc((void **)&d_x, n * sizeof(T)));	
    CUDA_SAFE_CALL(cudaMemcpy(d_b, h_b, n * sizeof(T), cudaMemcpyHostToDevice));
    CUDA_SAFE_CALL(cudaMemcpy(d_x, h_x, n * sizeof(T), cudaMemcpyHostToDevice));

	/* allocate temporary vectors */
	cublasAlloc(n * (m + 1), sizeof(T), (void **)&V);
	cublasAlloc(n, sizeof(T), (void **)&r);
    T *U = (T *)malloc(m * (m + 1) / 2 * sizeof(T));
    T *y = (T *)malloc((m + 1) * sizeof(T));
    T *c = (T *)malloc(m * sizeof(T));
    T *s = (T *)malloc(m * sizeof(T));
    T **v = (T **)malloc((m + 1) * sizeof(T *));    
    
    for(int i = 0; i <= m; ++i)
    	v[i] = V + i * n;    
    
    T b_square = cublasXnrm2(n, d_b, 1);
    T error_norm_rel = error_rel * b_square;

    io = 0;
    do
    {
    	// "aussere Iteration
    	++io;    
        spmv_single_gpu<T>(&d_spm, r, d_x);
        cublasXaxpy(n, -1.0, d_b, 1, r, 1);
        precon_single_gpu<Matrix_T, T>(&d_spm, &d_precon, h_precon, r, r);

        beta = cublasXnrm2(n, r, 1);
        cublasXcopy(n, r, 1, v[0], 1);
        cublasXscal(n, 1.0 / beta, v[0], 1);

        y[0] = beta;
        j = 0;
        uij = 0;
        do
        { 
    		// innere Iteration j=0,...,m-1
            u0j = uij;
            spmv_single_gpu<T>(&d_spm, v[j + 1], v[j]);
            precon_single_gpu<Matrix_T, T>(&d_spm, &d_precon, h_precon, v[j + 1], v[j + 1]);

            for(int i = 0; i < j + 1; i++) 
                U[u0j + i] = cublasXdot(n, v[j + 1], 1, V + i * n, 1);
            for(int i = 0; i < j + 1; i++) 
                cublasXaxpy(n, -U[u0j + i], V + i * n, 1, v[j + 1], 1);

            h = cublasXnrm2(n, v[j + 1], 1);
            cublasXscal(n, 1.0 / h, v[j + 1], 1);           

            for(int i = 0; i < j; ++i)
            {
            	// rotiere neue Spalte
            	T tmp = c[i] * U[uij] - s[i] * U[uij + 1];
            	U[uij + 1] = s[i] * U[uij] + c[i] * U[uij + 1];
            	U[uij] = tmp;
            	++uij;
            }
            {
            	// berechne neue Rotation
            	rd = U[uij];
            	dd = sqrt(rd * rd + h * h);
            	c[j] = rd / dd;
            	s[j] = -h / dd;
            	U[uij] = dd;
            	++uij;
            }
            {
            	// rotiere rechte Seite y (vorher: y[j+1]=0)
            	y[j + 1] = s[j] * y[j];
            	y[j] = c[j] * y[j];
            }
            ++j;
            error_norm = fabs(y[j]);
            k = m * (io - 1) + j;
            if(!(k % SHOW_INTERVAL)) printf("SIGLE DEVICE: interation num: %d, error: %f\n", k, error_norm);
        }
    	while(j < m && error_norm >= error_norm_rel);
        {   
            // minimiere bzgl Y
            Xtpsv("U", "N", "N", j, U, y, 1);
            
            /* correct X */
            for(int i = 0; i < j; i++) 
                cublasXaxpy(n, -y[i], V + i * n, 1, d_x, 1);
    	}
    }
    while(error_norm >= error_norm_rel && k < SUPER_MAX_ITER); 

	stop_timer(timer_compu, &t_compu_ms);
    *iter = k;
	fprintf(fpDat, "The computation time of single GPU is: %f ms\n", t_compu_ms);
    fprintf(fpDat, "The iteration of single GPU is: %d\n", k);  

    double precon_flop = get_precon_flop<T>(h_precon);
    double Gflops = (t_compu_ms == 0.0) ? 0.0 : 1;
    //((2.0 * h_spm->nnzero + 7.0 * n + precon_flop + (k + 1) * (2.0 * h_spm->nnzero + 12.0 * n + precon_flop)) / t_compu_ms) / 1e6;
    fprintf(fpDat, "The compute GFLOPs for PGMRES: %2.2f\n", Gflops);

	/* transfer solution from device to host */
	status = cublasGetVector(n, sizeof(T), d_x, 1, h_x, 1);
	if(status != CUBLAS_STATUS_SUCCESS)
	{
		fprintf(stderr, "!!!! device access error - 'd_x'\n");
		exit(EXIT_FAILURE);
	}
	
	/* free temporary vectors */
    cublasFree(V);
    cublasFree(r);
    free(U);
	free(y);
	free(c);
	free(s);
	free(v);
	/* free device memory */
    d_dealloc_matrix<T>(&d_spm);  
    d_dealloc_precon<T>(&d_precon);    
    CUDA_SAFE_CALL(cudaFree(d_b));
    CUDA_SAFE_CALL(cudaFree(d_x));
	
	/* shutdown CUBLAS */
	status = cublasShutdown();
	if(status != CUBLAS_STATUS_SUCCESS)
	{
		fprintf(stderr, "!!!! shutdown error\n");
		exit(EXIT_FAILURE);
	}

    return sqrt(error_norm / b_square);
}

#endif /* __PGMRES_DEVICE_H__ */
