/*
 * CUDA for Sparse Solver
 * Author: Luo Li     li.luo@sub.siat.ac.cn
 */ 

#ifndef __PGMRES_HOST_H__
#define __PGMRES_HOST_H__

#include "blas.h"

template <class Matrix_T, class Precon_T, class T>
T pgmres_host(FILE *fpDat, Matrix_T *h_spm, Precon_T *h_precon, T *h_b, T *h_x, const int max_iter, const T error_rel, int *iter)
{
    int n;
    if(h_spm->nrow == h_spm->ncol) 
        n = h_spm->nrow;
    else
    {
        printf("PGMRES_HOST: Nonsquare matrix doesn't supported.\n");
        exit(EXIT_FAILURE);
    }

    T beta, h, rd, dd, error_norm;
    int j, io, uij, u0j, k;
    int m = GMRES_RESTART;

    /* allocate temporary vectors */
    T *V = (T *)malloc(n * (m + 1) * sizeof(T));
    T *r = (T *)malloc(n * sizeof(T));    
    T *U = (T *)malloc(m * (m + 1) / 2 * sizeof(T));
    T *y = (T *)malloc((m + 1) * sizeof(T));
    T *c = (T *)malloc(m * sizeof(T));
    T *s = (T *)malloc(m * sizeof(T));
    T **v = (T **)malloc((m + 1) * sizeof(T *));
    
    unsigned int timer_compu = 0;
    double t_compu_ms = 0.0;
    start_timer(&timer_compu);	
    
    for(int i = 0; i <= m; ++i)
    	v[i] = V + i * n;

    T b_square = Xnrm2(n, h_b, 1); 
    T error_norm_rel = error_rel * b_square;
    
    io = 0;
    do
    {
    	// "aussere Iteration
    	++io;
        spmv_cpu<T>(h_spm, h_x, r);
    	Xaxpy(n, -1.0, h_b, 1, r, 1);
        precon_cpu<Matrix_T, T>(h_spm, h_precon, r, r);

    	beta = Xnrm2(n, r, 1);
    	Xcopy(n, r, 1, v[0], 1);
    	Xscal(n, 1.0 / beta, v[0], 1);
    
    	y[0] = beta;
    	j = 0;
    	uij = 0;
    	do
    	{
    		// innere Iteration j=0,...,m-1
    		u0j = uij;
            spmv_cpu<T>(h_spm, v[j], v[j + 1]);
            precon_cpu<Matrix_T, T>(h_spm, h_precon, v[j + 1], v[j + 1]);
            
            for(int i = 0; i < j + 1; i++) 
                U[u0j + i] = Xdot(n, v[j + 1], 1, V + i * n, 1);
            for(int i = 0; i < j + 1; i++) 
                Xaxpy(n, -U[u0j + i], V + i * n, 1, v[j + 1], 1);
            
    		//Xgemv("T", n, j + 1, 1.0, V, n, v[j + 1], 1, 0., U + u0j, 1);
    		//Xgemv("N", n, j + 1, -1.0, V, n, U + u0j, 1, 1.0, v[j + 1], 1);

    		h = Xnrm2(n, v[j + 1], 1);
    		Xscal(n, 1.0 / h, v[j + 1], 1);

    		for(int i = 0; i < j; ++i)
    		{
    			// rotiere neue Spalte
    			T tmp = c[i] * U[uij] - s[i] * U[uij + 1];
    			U[uij + 1] = s[i] * U[uij] + c[i] * U[uij + 1];
    			U[uij] = tmp;
    			++uij;
    		}
    		{
    			// berechne neue Rotation
    			rd = U[uij];
    			dd = sqrt(rd * rd + h * h);
    			c[j] = rd / dd;
    			s[j] = -h / dd;
    			U[uij] = dd;
    			++uij;
    		}
    		{
    			// rotiere rechte Seite y (vorher: y[j+1]=0)
    			y[j + 1] = s[j] * y[j];
    			y[j] = c[j] * y[j];
    		}
    		++j;
            error_norm = fabs(y[j]);
            k = m * (io - 1) + j;
            if(!(k % SHOW_INTERVAL)) printf("HOST: interation num: %d, error: %f\n", k, error_norm);
    	}
    	while(j < m && error_norm >= error_norm_rel);
    	{
    		// minimiere bzgl Y
    		Xtpsv("U", "N", "N", j, U, y, 1);

    		// korrigiere X
            
            for(int i = 0; i < j; i++) 
                Xaxpy(n, -y[i], V + i * n, 1, h_x, 1);

    		//Xgemv("N", n, j, -1.0, V, n, y, 1, 1.0, h_x, 1);
    	}
    }
    while(error_norm >= error_norm_rel && k < SUPER_MAX_ITER);  

    stop_timer(timer_compu, &t_compu_ms);
    *iter = k;
    fprintf(fpDat, "The iteration of CPU is: %d\n", k);  
    
    double precon_flop = get_precon_flop<T>(h_precon);
    double Gflops = (t_compu_ms == 0.0) ? 0.0 : 1;
    //((2.0 * h_spm->nnzero + 7.0 * n + precon_flop + (k + 1) * (2.0 * h_spm->nnzero + 12.0 * n + precon_flop)) / t_compu_ms) / 1e6;
    fprintf(fpDat, "The compute GFLOPs for PGMRES: %2.2f\n", Gflops);
    
    /* free temporary vectors */    
    free(V);
    free(r);
    free(U);
    free(y);
    free(c);
    free(s);
    free(v);
      
    return sqrt(error_norm / b_square);
}

#endif /* __PGMRES_HOST_H__ */
