/*
 * CUDA for Sparse Solver
 * Author: Luo Li     li.luo@sub.siat.ac.cn
 */ 

#ifndef __SPMATRIX_TRIANGLE_H__
#define __SPMATRIX_TRIANGLE_H__

/* Use for SSOR2: ssor subtriangle, etc. 
   lower or upper part of block diag array */
template <class T>
struct spm_array
{
    int nspm;	 /* Number of array */
    int nrow;    /* Global matrix row number */
    int ncol;    /* Global matrix column number */
    int nnzero;  /* Global matrix nnzero number */  
    spmatrix<T> *spm_entry; /* Data struct of each array element is a spmatrix sets */
};

/* CSC format for lower or upper triangle */
template <class T>
struct csc_triangle
{
    int nrow; 		  /* 1 less than the global matrix */
    int ncol; 		  /* 1 less than the global matrix */
    int nnzero; 	  /* nnzero number of this triangle */
    int nrow_per_col; /* Use in precon_ssor_single_gpu to compute dimBlock and dimGrid */
	///////////////CSC format
    int *col_ptr;     /* Column pointer array */
    int *row_ind;	  /* row index array */
    T *nz_val;		  /* nnzero val array */
};

template <class T>
struct csc_tri_array
{
    int ntri;	 /* Number of array */
    int nrow;    /* Global matrix row number */
    int ncol;    /* Global matrix column number */
    int nnzero;  /* Global matrix nnzero number */        
    csc_triangle<T> *csc_tri_entry; /* Data struct of each array element is a csc_triangle sets */
};

///////////////////////////////////////

template <class T>
int cmp_csc(const void *e1, const void *e2) 
{
    if (((nzinfo<T> *)e1)->col_num < ((nzinfo<T> *)e2)->col_num) return -1;
    if (((nzinfo<T> *)e1)->col_num > ((nzinfo<T> *)e2)->col_num) return 1;
    if (((nzinfo<T> *)e1)->row_num < ((nzinfo<T> *)e2)->row_num) return -1;
    if (((nzinfo<T> *)e1)->row_num > ((nzinfo<T> *)e2)->row_num) return -1;
    return 0;
}

template <class T>
int cmp_csc_reverse(const void *e1, const void *e2) 
{
    if (((nzinfo<T> *)e1)->col_num > ((nzinfo<T> *)e2)->col_num) return -1;
    if (((nzinfo<T> *)e1)->col_num < ((nzinfo<T> *)e2)->col_num) return 1;
    if (((nzinfo<T> *)e1)->row_num > ((nzinfo<T> *)e2)->row_num) return -1;
    if (((nzinfo<T> *)e1)->row_num < ((nzinfo<T> *)e2)->row_num) return 1;
    return 0;
}

/* Get the lower part of a matrix from COO(spmatrix), the lower part is also COO(spmatrix) form */
template <class T>
void get_lower(spmatrix<T> *spm, spmatrix<T> *spm_lower)
{
    if(spm->nrow != spm->ncol) 
    {
        printf("GET_LOWER: Nonsquare matrix doesn't supported.\n");
        exit(EXIT_FAILURE);
    }
    
    int row, colomn;
	/* The pure lower part that skims off the diag, so the row/column number is 1 less than the global matrix */
    spm_lower->nrow = spm->nrow - 1; 
    spm_lower->ncol = spm->ncol - 1; 
	/* Use 1 for to count nnzero for lower part */
    spm_lower->nnzero = 0;
    for(int i = 0; i < spm->nnzero; i++) 
    {
        row = (spm->nzentry)[i].row_num;
        colomn = (spm->nzentry)[i].col_num;
        if(row > colomn)
            spm_lower->nnzero++;
    }

    spm_lower->nnz_per_col = (int *)malloc(sizeof(int) * spm_lower->ncol);
    spm_lower->nzentry = (nzinfo<T> *)malloc(sizeof(nzinfo<T>) * spm_lower->nnzero);

    int j = 0;
    for(int i = 0; i < spm->nnzero; i++) 
    {
        row = (spm->nzentry)[i].row_num;
        colomn = (spm->nzentry)[i].col_num;
		/* For pure lower part */
        if(row > colomn)
        {
            (spm_lower->nnz_per_col)[colomn]++;
            (spm_lower->nzentry)[j] = (spm->nzentry)[i];
            j++;
        }
    }
}

template <class T>
void get_lower(spmatrix<T> *spm, spm_array<T> *spm_lower)
{
    int n;
    if(spm->nrow == spm->ncol) 
        n = spm->nrow;
    else
    {
        printf("GET_LOWER: Nonsquare matrix doesn't supported.\n");
        exit(EXIT_FAILURE);
    }
    spm_lower->nrow = n;
    spm_lower->ncol = n;
    spm_lower->nnzero = spm->nnzero;

    int dimBlock = DIMBLOCK;
    spm_lower->nspm = (n + dimBlock - 1) / dimBlock;
    spm_lower->spm_entry = (spmatrix<T> *)malloc(sizeof(spmatrix<T>) * spm_lower->nspm);

    int offset;
    for(int i = 0; i < spm_lower->nspm; i++)
    {
        offset = i * dimBlock;
        (spm_lower->spm_entry)[i].nrow = min(dimBlock, n - offset) - 1;
        (spm_lower->spm_entry)[i].ncol = (spm_lower->spm_entry)[i].nrow;
        (spm_lower->spm_entry)[i].nnzero = 0;
        (spm_lower->spm_entry)[i].nnz_per_col = (int *)malloc(sizeof(int) * (spm_lower->spm_entry)[i].ncol);
    }

    int row, colomn, block_row, block_col, row_in_block, col_in_block;
    for(int i = 0; i < spm->nnzero; i++) 
    {
        row = (spm->nzentry)[i].row_num;
        colomn = (spm->nzentry)[i].col_num;
        block_row = row / dimBlock;
        block_col = colomn / dimBlock;
        row_in_block = row - block_row * dimBlock;
        col_in_block = colomn - block_col * dimBlock;
        
        if(block_row == block_col && row_in_block > col_in_block)
            (spm_lower->spm_entry)[block_row].nnzero++;
    }
    
    for(int i = 0; i < spm_lower->nspm; i++)
        (spm_lower->spm_entry)[i].nzentry = (nzinfo<T> *)malloc(sizeof(nzinfo<T>) * (spm_lower->spm_entry)[i].nnzero);

    int *ind = (int *)malloc(sizeof(int) * spm_lower->nspm);
    memset(ind, 0, spm_lower->nspm * sizeof(int));
    for(int i = 0; i < spm->nnzero; i++) 
    {
        row = (spm->nzentry)[i].row_num;
        colomn = (spm->nzentry)[i].col_num;
        block_row = row / dimBlock;//block coordinate
        block_col = colomn / dimBlock;
        row_in_block = row - block_row * dimBlock;//relative coordinate in block
        col_in_block = colomn - block_col * dimBlock;  
 
        if(block_row == block_col && row_in_block > col_in_block)
        {
            ((spm_lower->spm_entry)[block_row].nnz_per_col)[col_in_block]++;
            nzinfo<T> *temp = &((spm_lower->spm_entry)[block_row].nzentry)[ind[block_row]];
            temp->val = (spm->nzentry)[i].val;
            temp->row_num = row_in_block;
            temp->col_num = col_in_block;
            ind[block_row]++;
        }
    }
    free(ind);
}

/* Get the upper part of a matrix from COO(spmatrix), the upper part is also COO(spmatrix) form */
template <class T>
void get_upper(spmatrix<T> *spm, spmatrix<T> *spm_upper)
{
    if(spm->nrow != spm->ncol) 
    {
        printf("GET_UPPER: Nonsquare matrix doesn't supported.\n");
        exit(EXIT_FAILURE);
    }
    
    int row, colomn;
	/* The pure upper part that skims off the diag, so the row/column number is 1 less than the global matrix */
    spm_upper->nrow = spm->nrow - 1; 
    spm_upper->ncol = spm->ncol - 1; 
	/* Use 1 for to count nnzero for upper part */
    spm_upper->nnzero = 0;
    for(int i = 0; i < spm->nnzero; i++) 
    {
        row = (spm->nzentry)[i].row_num;
        colomn = (spm->nzentry)[i].col_num;
        if(row < colomn)
            spm_upper->nnzero++;
    }

    spm_upper->nnz_per_col = (int *)malloc(sizeof(int) * spm_upper->ncol);
    spm_upper->nzentry = (nzinfo<T> *)malloc(sizeof(nzinfo<T>) * spm_upper->nnzero);
    
    int j = 0;
    for(int i = 0; i < spm->nnzero; i++) 
    {
        row = (spm->nzentry)[i].row_num;
        colomn = (spm->nzentry)[i].col_num;
		/* For pure upper part */
        if(row < colomn)
        {
            (spm_upper->nnz_per_col)[colomn]++;
            (spm_upper->nzentry)[j] = (spm->nzentry)[i];
            j++;
        }
    }   
}

template <class T>
void get_upper(spmatrix<T> *spm, spm_array<T> *spm_upper)
{
    int n;
    if(spm->nrow == spm->ncol) 
        n = spm->nrow;
    else
    {
        printf("GET_UPPER: Nonsquare matrix doesn't supported.\n");
        exit(EXIT_FAILURE);
    }
    spm_upper->nrow = n;
    spm_upper->ncol = n;
    spm_upper->nnzero = spm->nnzero;
    
    int dimBlock = DIMBLOCK;
    spm_upper->nspm = (n + dimBlock - 1) / dimBlock;
    spm_upper->spm_entry = (spmatrix<T> *)malloc(sizeof(spmatrix<T>) * spm_upper->nspm);

    int offset;
    for(int i = 0; i < spm_upper->nspm; i++)
    {
        offset = i * dimBlock;
        (spm_upper->spm_entry)[i].nrow = min(dimBlock, n - offset) - 1;
        (spm_upper->spm_entry)[i].ncol = (spm_upper->spm_entry)[i].nrow;
        (spm_upper->spm_entry)[i].nnzero = 0;
        (spm_upper->spm_entry)[i].nnz_per_col = (int *)malloc(sizeof(int) * (spm_upper->spm_entry)[i].ncol);
    }

    int row, colomn, block_row, block_col, row_in_block, col_in_block;
    int row_bw, col_bw;
    for(int i = spm->nnzero - 1; i >= 0; i--) 
    {
        row = (spm->nzentry)[i].row_num;
        colomn = (spm->nzentry)[i].col_num;
        row_bw = n - 1 - row;//backward id
        col_bw = n - 1 - colomn; 
        block_row = row_bw / dimBlock;
        block_col = col_bw / dimBlock;
        row_in_block = row_bw - block_row * dimBlock;
        col_in_block = col_bw - block_col * dimBlock;
        
        if(block_row == block_col && row_in_block > col_in_block)//">" because backward 
            (spm_upper->spm_entry)[block_row].nnzero++;
    }
    
    for(int i = 0; i < spm_upper->nspm; i++)
        (spm_upper->spm_entry)[i].nzentry = (nzinfo<T> *)malloc(sizeof(nzinfo<T>) * (spm_upper->spm_entry)[i].nnzero);

    int *ind = (int *)malloc(sizeof(int) * spm_upper->nspm);
    memset(ind, 0, spm_upper->nspm * sizeof(int));    
    for(int i = spm->nnzero - 1; i >= 0; i--) 
    {
        row = (spm->nzentry)[i].row_num;
        colomn = (spm->nzentry)[i].col_num;
        row_bw = n - 1 - row;
        col_bw = n - 1 - colomn; 
        block_row = row_bw / dimBlock;
        block_col = col_bw / dimBlock;
        row_in_block = row_bw - block_row * dimBlock;
        col_in_block = col_bw - block_col * dimBlock;        
 
        if(block_row == block_col && row_in_block > col_in_block)
        {
            ((spm_upper->spm_entry)[block_row].nnz_per_col)[col_in_block]++;
            nzinfo<T> *temp = &((spm_upper->spm_entry)[block_row].nzentry)[ind[block_row]];
            temp->val = (spm->nzentry)[i].val;
            temp->row_num = row_in_block;
            temp->col_num = col_in_block;            
            ind[block_row]++;            
        }
    }
    free(ind);
}

template <class T>
void clear_triangle(spmatrix<T> *spm_tri) 
{
    free(spm_tri->nnz_per_col);    
    free(spm_tri->nzentry);
}

template <class T>
void clear_triangle(spm_array<T> *spm_tri) 
{
    for(int i = 0; i < spm_tri->nspm; i++)
       clear_triangle<T>(&(spm_tri->spm_entry)[i]);
    free(spm_tri->spm_entry);
}

/* Convert the matrix from COO to CSC format */
template <class T>
void convert_lower(spmatrix<T> *spm_lower, csc_triangle<T> *csc_lower)
{
	/* !!! Resort the data to be column-major, notice that spm_lower->nzentry is changed forever */
    qsort(spm_lower->nzentry, spm_lower->nnzero, sizeof(nzinfo<T>), cmp_csc<T>);

    csc_lower->nrow = spm_lower->nrow; 
    csc_lower->ncol = spm_lower->ncol; 
    csc_lower->nnzero = spm_lower->nnzero;

	/* Compute the maximum of number of nnzero among all colunms, 
	max_nnz_per_col is used for computing dimGrid that needed in CUDA kernels */
    int max_nnz_per_col = 0;
    for(int i = 0; i < csc_lower->ncol; i++) 
        max_nnz_per_col = max(max_nnz_per_col, (spm_lower->nnz_per_col)[i]);
    csc_lower->nrow_per_col = max_nnz_per_col;
    
	/* Strategy for aligned memory access optimization. Padding 0 to each column to make it times of 16. 
	   Seldom use. */
    #if USE_ALIGNED_OPT_PRECON
    csc_lower->col_ptr = (int *)malloc(sizeof(int) * (csc_lower->ncol + 1));
	/* Handle the case when the first nnzero is not at the first column, 
	that (spm_lower->nzentry)[0].col_num may not be 0, hence col_ptr of those columns is 0.	*/
    int curcol, prevcol = (spm_lower->nzentry)[0].col_num;
    for(int tmp_col = 0; tmp_col <= prevcol; tmp_col++)
        (csc_lower->col_ptr)[tmp_col] = 0;
    
    for(int i = 1; i < csc_lower->nnzero; i++) 
    {
        curcol = (spm_lower->nzentry)[i].col_num;
        if(curcol != prevcol && curcol != (prevcol + 1)) 
        {
			/* Handle the case when there is no nnzero at those columns, 
			but i now already stands for a nnzero at some column behind.
			col_ptr of those columns are also set to be i. 
			
			1. diag element is 0, while some other rows of this column is not 0.
			2. diag element is not 0, while all other rows of this column are 0.
			3. diag element is 0, and all other rows of this column are 0.
			All three cases above result in precon_ssor_lower functions, where when for(int k = 0; k < n - 1; k++)
			traversing the current of certain case. 
			case 1 seldom happens, the result may not be right.
			case 2 and 3 are similar, 	
			start = col_ptr[k];
			dist = col_ptr[k + 1];
			are the same so that nothing of x would be modified. 
			*/
            for(int tmp_col = prevcol + 1; tmp_col <= curcol; tmp_col++)
                (csc_lower->col_ptr)[tmp_col] = i;
            prevcol = curcol;
        }
        else if(curcol == (prevcol + 1))
        {
            prevcol = curcol;
            (csc_lower->col_ptr)[curcol] = i;
        }
    }
    if(csc_lower->ncol != (prevcol + 1)) 
    {
        for(int tmp_col = prevcol + 1; tmp_col < csc_lower->ncol; tmp_col++)
            (csc_lower->col_ptr)[tmp_col] = csc_lower->nnzero;        
    }
    (csc_lower->col_ptr)[csc_lower->ncol] = csc_lower->nnzero;
    
	/* pad_nnzero refers to final nnzero after padded */
    int nnz_col, pad_nnzero = 0;
    for(int i = 0; i < csc_lower->ncol; i++) 
    {
        nnz_col = (csc_lower->col_ptr)[i + 1] - (csc_lower->col_ptr)[i];
        if(nnz_col % HALFWARP) 
            pad_nnzero += nnz_col + HALFWARP - (nnz_col % HALFWARP); 
        else 
            pad_nnzero += nnz_col;
    }

    free(csc_lower->col_ptr);
    #if USE_PAGE_LOCKED
    CUDA_SAFE_CALL(cudaMallocHost((void **)&csc_lower->col_ptr, (csc_lower->ncol + 1) * sizeof(int)));
    CUDA_SAFE_CALL(cudaMallocHost((void **)&csc_lower->row_ind, pad_nnzero * sizeof(int)));
    CUDA_SAFE_CALL(cudaMallocHost((void **)&csc_lower->nz_val, pad_nnzero * sizeof(T)));
    #else
    csc_lower->col_ptr = (int *)malloc(sizeof(int) * (csc_lower->ncol + 1));
    csc_lower->row_ind = (int *)malloc(sizeof(int) * pad_nnzero);
    csc_lower->nz_val = (T *)malloc(sizeof(T) * pad_nnzero);
    #endif
    /* The same as above */
    int cur_col, prev_col = (spm_lower->nzentry)[0].col_num;
    for(int tmp_col = 0; tmp_col <= prev_col; tmp_col++)
        (csc_lower->col_ptr)[tmp_col] = 0;

    int cur_pad_nnzero = 0;    
    for(int i = 0; i < csc_lower->nnzero; i++) 
    {
        cur_col = (spm_lower->nzentry)[i].col_num;
        int cur_mode = cur_pad_nnzero % HALFWARP;
        if(cur_col != prev_col && cur_col != (prev_col + 1)) 
        {
			/* Pad 0 */
            if(cur_mode)
            {
                int fillCount = HALFWARP - cur_mode;
                for(int j = 0; j < fillCount; j++) 
                {
                    (csc_lower->row_ind)[cur_pad_nnzero] = -1;
                    (csc_lower->nz_val)[cur_pad_nnzero] = 0.0;
                    cur_pad_nnzero++;
                }                
            }
			/* Handle the case when there is no nnzero at those columns, 
			but cur_pad_nnzero now already stands for a nnzero at some column behind.
			col_ptr of those columns are also set to be cur_pad_nnzero. */
            for(int tmp_col = prev_col + 1; tmp_col <= cur_col; tmp_col++)
                (csc_lower->col_ptr)[tmp_col] = cur_pad_nnzero;
            prev_col = cur_col;
        }
        else if(cur_col == (prev_col + 1))
        {
            if(cur_mode)
            {
                int fillCount = HALFWARP - cur_mode;
                for(int j = 0; j < fillCount; j++) 
                {
                    (csc_lower->row_ind)[cur_pad_nnzero] = -1;
                    (csc_lower->nz_val)[cur_pad_nnzero] = 0.0;
                    cur_pad_nnzero++;
                }
            }
            prev_col = cur_col;
            (csc_lower->col_ptr)[cur_col] = cur_pad_nnzero;
        }
        (csc_lower->row_ind)[cur_pad_nnzero] = (spm_lower->nzentry)[i].row_num;
        (csc_lower->nz_val)[cur_pad_nnzero] = (spm_lower->nzentry)[i].val;
        cur_pad_nnzero++;
    }
    int cur_mode = cur_pad_nnzero % HALFWARP;
    if(cur_mode)
    {
        int fillCount = HALFWARP - cur_mode;
        for(int j = 0; j < fillCount; j++) 
        {
            (csc_lower->row_ind)[cur_pad_nnzero] = -1;
            (csc_lower->nz_val)[cur_pad_nnzero] = 0.0;
            cur_pad_nnzero++;
        }
    }
	/* If there is no nnzeros at the last several columns, same strategy as above, equal to pad_nnzero. */
    if(csc_lower->ncol != (prev_col + 1))
    {
        for(int tmp_col = prev_col + 1; tmp_col < csc_lower->ncol; tmp_col++)
            (csc_lower->col_ptr)[tmp_col] = pad_nnzero;        
    }
    (csc_lower->col_ptr)[csc_lower->ncol] = pad_nnzero;
    csc_lower->nnzero = pad_nnzero;
    #else
    #if USE_PAGE_LOCKED
    CUDA_SAFE_CALL(cudaMallocHost((void **)&csc_lower->col_ptr, (csc_lower->ncol + 1) * sizeof(int)));
    CUDA_SAFE_CALL(cudaMallocHost((void **)&csc_lower->row_ind, csc_lower->nnzero * sizeof(int)));
    CUDA_SAFE_CALL(cudaMallocHost((void **)&csc_lower->nz_val, csc_lower->nnzero * sizeof(T)));
    #else
    csc_lower->col_ptr = (int *)malloc(sizeof(int) * (csc_lower->ncol + 1));
    csc_lower->row_ind = (int *)malloc(sizeof(int) * csc_lower->nnzero);
    csc_lower->nz_val = (T *)malloc(sizeof(T) * csc_lower->nnzero);
    #endif

    int cur_col, prev_col = (spm_lower->nzentry)[0].col_num;
	/* Handle the case when the first nnzero is not at the first column, 
	that (spm_lower->nzentry)[0].col_num may not be 0, hence col_ptr of those columns is 0.	*/
    for(int tmp_col = 0; tmp_col <= prev_col; tmp_col++)
        (csc_lower->col_ptr)[tmp_col] = 0;
    
    for(int i = 1; i < csc_lower->nnzero; i++) 
    {
        cur_col = (spm_lower->nzentry)[i].col_num;
        if(cur_col != prev_col && cur_col != (prev_col + 1)) 
        {
			/* Handle the case when there is no nnzero at those columns, 
			but i now already stands for a nnzero at some column behind.
			col_ptr of those columns are also set to be i. 
			
			1. diag element is 0, while some other rows of this column is not 0.
			2. diag element is not 0, while all other rows of this column are 0.
			3. diag element is 0, and all other rows of this column are 0.
			All three cases above result in precon_ssor_lower functions, where when for(int k = 0; k < n - 1; k++)
			traversing the current of certain case. 
			case 1 seldom happens, the result may not be right.
			case 2 and 3 are similar, 	
			start = col_ptr[k];
			dist = col_ptr[k + 1];
			are the same so that nothing of x would be modified. 
			*/
            for(int tmp_col = prev_col + 1; tmp_col <= cur_col; tmp_col++)
                (csc_lower->col_ptr)[tmp_col] = i;
            prev_col = cur_col;
        }
        else if(cur_col == (prev_col + 1))
        {
            prev_col = cur_col;
			/* i now already stands for a nnzero at some column behind. */
            (csc_lower->col_ptr)[cur_col] = i;
        }
    }
	/* If there is no nnzeros at the last several columns, same strategy as above, equal to csc_lower->nnzero. */
    if(csc_lower->ncol != (prev_col + 1)) 
    {
        for(int tmp_col = prev_col + 1; tmp_col < csc_lower->ncol; tmp_col++)
            (csc_lower->col_ptr)[tmp_col] = csc_lower->nnzero;        
    }
    (csc_lower->col_ptr)[csc_lower->ncol] = csc_lower->nnzero;
    ///////////col_ptr array above, row_ind and nz_val below.
    for(int i = 0; i < csc_lower->nnzero; i++) 
    {
        (csc_lower->row_ind)[i] = (spm_lower->nzentry)[i].row_num;
        (csc_lower->nz_val)[i] = (spm_lower->nzentry)[i].val;
    }    
    #endif
}

template <class T>
void convert_lower(spm_array<T> *spm_lower, csc_tri_array<T> *csc_lower)
{
    csc_lower->ntri = spm_lower->nspm;
    csc_lower->nrow = spm_lower->nrow;
    csc_lower->ncol = spm_lower->ncol;
    csc_lower->nnzero = spm_lower->nnzero;

    csc_lower->csc_tri_entry = (csc_triangle<T> *)malloc(sizeof(csc_triangle<T>) * csc_lower->ntri);
    for(int i = 0; i < csc_lower->ntri; i++)
        convert_lower<T>(&(spm_lower->spm_entry)[i], &(csc_lower->csc_tri_entry)[i]);
}

template <class T>
void convert_upper(spmatrix<T> *spm_upper, csc_triangle<T> *csc_upper)
{
    qsort(spm_upper->nzentry, spm_upper->nnzero, sizeof(nzinfo<T>), cmp_csc_reverse<T>);

    csc_upper->nrow = spm_upper->nrow; 
    csc_upper->ncol = spm_upper->ncol; 
    csc_upper->nnzero = spm_upper->nnzero;

    int max_nnz_per_col = 0;
    for(int i = 0; i < csc_upper->ncol; i++) 
        max_nnz_per_col = max(max_nnz_per_col, (spm_upper->nnz_per_col)[i]);
    csc_upper->nrow_per_col = max_nnz_per_col;    

    #if USE_ALIGNED_OPT_PRECON
    csc_upper->col_ptr = (int *)malloc(sizeof(int) * (csc_upper->ncol + 1));

    int curcol, colptr_num, prevcol = (spm_upper->nzentry)[0].col_num - 1;
    for(int tmp_col = 0; tmp_col <= csc_upper->ncol - 1 - prevcol; tmp_col++)
        (csc_upper->col_ptr)[tmp_col] = 0;
    
    for(int i = 1; i < csc_upper->nnzero; i++) 
    {
        curcol = (spm_upper->nzentry)[i].col_num - 1;
        if(curcol != prevcol && curcol != (prevcol - 1)) 
        {
            for(int tmp_col = prevcol - 1; tmp_col > curcol; tmp_col--)
                (csc_upper->col_ptr)[csc_upper->ncol - 1 - tmp_col] = i;
            prevcol = curcol;
            colptr_num = csc_upper->ncol - 1 - curcol;
            (csc_upper->col_ptr)[colptr_num] = i;            
        }
        else if(curcol == (prevcol - 1))        
        {
            prevcol = curcol;
            colptr_num = csc_upper->ncol - 1 - curcol;
            (csc_upper->col_ptr)[colptr_num] = i;
        }
    }
    if(-1 != (prevcol - 1)) 
    {
        for(int tmp_col = prevcol - 1; tmp_col > -1; tmp_col--)
            (csc_upper->col_ptr)[csc_upper->ncol - 1 - tmp_col] = csc_upper->nnzero;        
    }
    (csc_upper->col_ptr)[csc_upper->ncol] = csc_upper->nnzero;
    
    int nnz_col, pad_nnzero = 0;
    for(int i = 0; i < csc_upper->ncol; i++) 
    {
        nnz_col = (csc_upper->col_ptr)[i + 1] - (csc_upper->col_ptr)[i];
        if(nnz_col % HALFWARP) 
            pad_nnzero += nnz_col + HALFWARP - (nnz_col % HALFWARP); 
        else 
            pad_nnzero += nnz_col;
    }

    free(csc_upper->col_ptr);
    #if USE_PAGE_LOCKED
    CUDA_SAFE_CALL(cudaMallocHost((void **)&csc_upper->col_ptr, (csc_upper->ncol + 1) * sizeof(int)));
    CUDA_SAFE_CALL(cudaMallocHost((void **)&csc_upper->row_ind, pad_nnzero * sizeof(int)));
    CUDA_SAFE_CALL(cudaMallocHost((void **)&csc_upper->nz_val, pad_nnzero * sizeof(T)));
    #else
    csc_upper->col_ptr = (int *)malloc(sizeof(int) * (csc_upper->ncol + 1));
    csc_upper->row_ind = (int *)malloc(sizeof(int) * pad_nnzero);
    csc_upper->nz_val = (T *)malloc(sizeof(T) * pad_nnzero);
    #endif
    
    int cur_col, col_ptr_num, prev_col = (spm_upper->nzentry)[0].col_num - 1;
    for(int tmp_col = 0; tmp_col <= csc_upper->ncol - 1 - prev_col; tmp_col++)
        (csc_upper->col_ptr)[tmp_col] = 0;

    int cur_pad_nnzero = 0;    
    for(int i = 0; i < csc_upper->nnzero; i++) 
    {
        cur_col = (spm_upper->nzentry)[i].col_num - 1;
        int cur_mode = cur_pad_nnzero % HALFWARP;
        if(cur_col != prev_col && cur_col != (prev_col - 1)) 
        {
            if(cur_mode)
            {
                int fillCount = HALFWARP - cur_mode;
                for(int j = 0; j < fillCount; j++) 
                {
                    (csc_upper->row_ind)[cur_pad_nnzero] = -1;
                    (csc_upper->nz_val)[cur_pad_nnzero] = 0.0;
                    cur_pad_nnzero++;
                }                
            }            
            for(int tmp_col = prev_col - 1; tmp_col >= cur_col; tmp_col--)
                (csc_upper->col_ptr)[csc_upper->ncol - 1 - tmp_col] = cur_pad_nnzero;
            prev_col = cur_col;           
        }
        else if(cur_col == (prev_col - 1))        
        {
            if(cur_mode)
            {
                int fillCount = HALFWARP - cur_mode;
                for(int j = 0; j < fillCount; j++) 
                {
                    (csc_upper->row_ind)[cur_pad_nnzero] = -1;
                    (csc_upper->nz_val)[cur_pad_nnzero] = 0.0;
                    cur_pad_nnzero++;
                }                
            }            
            prev_col = cur_col;
            col_ptr_num = csc_upper->ncol - 1 - cur_col;
            (csc_upper->col_ptr)[col_ptr_num] = cur_pad_nnzero;
        }
        (csc_upper->row_ind)[cur_pad_nnzero] = (spm_upper->nzentry)[i].row_num;
        (csc_upper->nz_val)[cur_pad_nnzero] = (spm_upper->nzentry)[i].val;
        cur_pad_nnzero++;
    }
    int cur_mode = cur_pad_nnzero % HALFWARP;
    if(cur_mode)
    {
        int fillCount = HALFWARP - cur_mode;
        for(int j = 0; j < fillCount; j++) 
        {
            (csc_upper->row_ind)[cur_pad_nnzero] = -1;
            (csc_upper->nz_val)[cur_pad_nnzero] = 0.0;
            cur_pad_nnzero++;
        }                
    }
    if(-1 != (prev_col - 1)) 
    {
        for(int tmp_col = prev_col - 1; tmp_col > -1; tmp_col--)
            (csc_upper->col_ptr)[csc_upper->ncol - 1 - tmp_col] = pad_nnzero;        
    }
    (csc_upper->col_ptr)[csc_upper->ncol] = pad_nnzero;
    csc_upper->nnzero = pad_nnzero;
    #else
    #if USE_PAGE_LOCKED
    CUDA_SAFE_CALL(cudaMallocHost((void **)&csc_upper->col_ptr, (csc_upper->ncol + 1) * sizeof(int)));
    CUDA_SAFE_CALL(cudaMallocHost((void **)&csc_upper->row_ind, csc_upper->nnzero * sizeof(int)));
    CUDA_SAFE_CALL(cudaMallocHost((void **)&csc_upper->nz_val, csc_upper->nnzero * sizeof(T)));
    #else
    csc_upper->col_ptr = (int *)malloc(sizeof(int) * (csc_upper->ncol + 1));
    csc_upper->row_ind = (int *)malloc(sizeof(int) * csc_upper->nnzero);
    csc_upper->nz_val = (T *)malloc(sizeof(T) * csc_upper->nnzero);
    #endif
    
    int cur_col, col_ptr_num, prev_col = (spm_upper->nzentry)[0].col_num - 1;
    for(int tmp_col = 0; tmp_col <= csc_upper->ncol - 1 - prev_col; tmp_col++)
        (csc_upper->col_ptr)[tmp_col] = 0;

    for(int i = 1; i < csc_upper->nnzero; i++) 
    {
        cur_col = (spm_upper->nzentry)[i].col_num - 1;
        if(cur_col != prev_col && cur_col != (prev_col - 1)) 
        {
            for(int tmp_col = prev_col - 1; tmp_col >= cur_col; tmp_col--)
                (csc_upper->col_ptr)[csc_upper->ncol - 1 - tmp_col] = i;
            prev_col = cur_col;           
        }
        else if(cur_col == (prev_col - 1))        
        {
            prev_col = cur_col;
            col_ptr_num = csc_upper->ncol - 1 - cur_col;
            (csc_upper->col_ptr)[col_ptr_num] = i;
        }
    }
    if(-1 != (prev_col - 1)) 
    {
        for(int tmp_col = prev_col - 1; tmp_col > -1; tmp_col--)
            (csc_upper->col_ptr)[csc_upper->ncol - 1 - tmp_col] = csc_upper->nnzero;        
    }
    (csc_upper->col_ptr)[csc_upper->ncol] = csc_upper->nnzero;
    
    for(int i = 0; i < csc_upper->nnzero; i++) 
    {
        (csc_upper->row_ind)[i] = (spm_upper->nzentry)[i].row_num;
        (csc_upper->nz_val)[i] = (spm_upper->nzentry)[i].val;
    }    
    #endif
}

template <class T>
void convert_upper(spm_array<T> *spm_upper, csc_tri_array<T> *csc_upper)
{
    csc_upper->ntri = spm_upper->nspm;
    csc_upper->nrow = spm_upper->nrow;
    csc_upper->ncol = spm_upper->ncol;
    csc_upper->nnzero = spm_upper->nnzero;
    
    csc_upper->csc_tri_entry = (csc_triangle<T> *)malloc(sizeof(csc_triangle<T>) * csc_upper->ntri);
    for(int i = 0; i < csc_upper->ntri; i++)
        convert_lower<T>(&(spm_upper->spm_entry)[i], &(csc_upper->csc_tri_entry)[i]);
}

template <class T>
void clear_csc_triangle(csc_triangle<T> *csc_tri) 
{
    #if USE_PAGE_LOCKED
    CUDA_SAFE_CALL(cudaFreeHost(csc_tri->col_ptr));
    CUDA_SAFE_CALL(cudaFreeHost(csc_tri->row_ind));
    CUDA_SAFE_CALL(cudaFreeHost(csc_tri->nz_val));
    #else
    free(csc_tri->col_ptr);
    free(csc_tri->row_ind);
    free(csc_tri->nz_val);
    #endif
}

template <class T>
void clear_csc_triangle(csc_tri_array<T> *csc_tri) 
{
    for(int i = 0; i < csc_tri->ntri; i++)
        clear_csc_triangle<T>(&(csc_tri->csc_tri_entry)[i]);
    free(csc_tri->csc_tri_entry);
}

template <class T>
void print_triangle(spmatrix<T> *spm_tri)
{
    printf("Triangle of spm: \n");
    int row, colomn;
    T val;
    for(int i = 0; i < spm_tri->nnzero; i++) 
    {
        row = (spm_tri->nzentry)[i].row_num;
        colomn = (spm_tri->nzentry)[i].col_num;
        val = (spm_tri->nzentry)[i].val;
        printf("%d, %d, %f\n", row, colomn, val);
    }
    printf("\n\n");
}

template <class T>
void print_triangle(spm_array<T> *spm_tri)
{
    printf("Subtriangle of spm: \n");    
    for(int i = 0; i < spm_tri->nspm; i++)
    {
        printf("The %d subtriangle: \n", i); 
        print_triangle<T>(&(spm_tri->spm_entry)[i]);
    }
}

template <class T>
void print_triangle(csc_triangle<T> *csc_tri)
{
    printf("Triangle of CSC format: \n");
    for(int i = 0; i < csc_tri->ncol; i++) 
    {
        printf("%d col: ", i);
        for(int jj = (csc_tri->col_ptr)[i]; jj < (csc_tri->col_ptr)[i + 1]; jj++) 
            printf("(%d %f), ", (csc_tri->row_ind)[jj], (csc_tri->nz_val)[jj]);
        printf("\n");
    }
    printf("\n\n");
}

template <class T>
void print_triangle(csc_tri_array<T> *csc_tri)
{
    printf("Subtriangle of CSC format: \n");
    for(int i = 0; i < csc_tri->ntri; i++)
    {
        printf("The %d subtriangle: \n", i); 
        print_triangle<T>(&(csc_tri->csc_tri_entry)[i]);
    }
}

#endif /* __SPMATRIX_TRIANGLE_H__ */
