/*
 * CUDA for Sparse Solver
 * Author: Luo Li     li.luo@sub.siat.ac.cn
 */ 

#ifndef __SPMATRIX_H__
#define __SPMATRIX_H__

#include "mmio.h"

template <class T>
struct nzinfo
{
    int row_num;
    int col_num;
    T val;
};

template <class T>
struct spmatrix
{
    int nrow;				/* matrix row number */
    int ncol;				/* matrix column number */
    int nnzero;				/* matrix nnzero number */
    int *nnz_per_row;		/* Array contains numbers of nnzero in each row. Use for CSR format */
    int *nnz_per_col; 		/* Array contains numbers of nnzero in each column. Use for CSC format */
    nzinfo<T> *nzentry;		/* Data struct(COO) of element */
};

template <class T>
spmatrix<T> read_spmatrix(char *mm_filename)
{
    FILE *fid = fopen(mm_filename, "r");
    if (fid == NULL)
    {
        printf("Unable to open file %s\n", mm_filename);
        exit(EXIT_FAILURE);
    }
    spmatrix<T> spm;
    MM_typecode matcode;
    
    if (mm_read_banner(fid, &matcode) != 0)
    {
        printf("Could not process Matrix Market banner\n");
        exit(EXIT_FAILURE);
    }
    if (!mm_is_valid(matcode)){
        printf("Invalid Matrix Market file.\n");
        exit(EXIT_FAILURE);
    }
    if (!((mm_is_real(matcode) || mm_is_integer(matcode) || mm_is_pattern(matcode)) && mm_is_coordinate(matcode) && mm_is_sparse(matcode) ) )
    {
        printf("Sorry, the application of such matrix does not support :");
        printf("Market Market type: [%s]\n", mm_typecode_to_str(matcode));
        printf("Only sparse real-valued or pattern coordinate matrices are supported\n");
        exit(EXIT_FAILURE);
    }
    int nrow, ncol, nnzero;
    if (mm_read_mtx_crd_size(fid, &nrow, &ncol, &nnzero) != 0)
        exit(EXIT_FAILURE);

    spm.nrow = nrow;
    spm.ncol = ncol;
    spm.nnzero = nnzero;
    spm.nnz_per_row = (int *)malloc(sizeof(int) * spm.nrow);
    memset(spm.nnz_per_row, 0, spm.nrow * sizeof(int));
    spm.nzentry = (nzinfo<T> *)malloc(sizeof(nzinfo<T>) * spm.nnzero);

    double temp;
    if (mm_is_pattern(matcode))
    {   // pattern matrix defines sparsity pattern, but not values
        for(int i = 0; i < spm.nnzero; i++)
        {
            if(fscanf(fid, "%d %d\n", &((spm.nzentry)[i].row_num), &((spm.nzentry)[i].col_num)) >=2 ) 
            {
                (spm.nzentry)[i].row_num--;//0 based  
                spm.nnz_per_row[(spm.nzentry)[i].row_num]++;
                (spm.nzentry)[i].col_num--;
                temp = 1.0;//use value 1.0 for all nonzero entries 
                (spm.nzentry)[i].val = static_cast<T>(temp);
            } 
        }
    }else if(mm_is_real(matcode) || mm_is_integer(matcode))
    {
        for(int i = 0; i < spm.nnzero; i++)
        {
			/* Notice that a double float should be scanf with lf format, some trouble happens if %f. */
            if (fscanf(fid, "%d %d %lf\n", &((spm.nzentry)[i].row_num), &((spm.nzentry)[i].col_num), &temp) >= 3) 
            {	
                (spm.nzentry)[i].row_num--;//0 based  
                spm.nnz_per_row[(spm.nzentry)[i].row_num]++;
                (spm.nzentry)[i].col_num--;                             
                (spm.nzentry)[i].val = static_cast<T>(temp);            	
            } 
        }
    }else 
    {
        printf("Unrecognized data type\n");
        exit(EXIT_FAILURE);
    }

    fclose(fid);

    /* if symmetric, reproduce the diagonals */
    if(mm_is_symmetric(matcode))
    {   //duplicate off diagonal entries
        int off_diagonals = 0;
        for(int i = 0; i < spm.nnzero; i++)
        {
            if((spm.nzentry)[i].row_num != (spm.nzentry)[i].col_num)
                off_diagonals++;
        }
        int ture_nnzero = 2 * off_diagonals + (spm.nnzero - off_diagonals);

        int *new_nnz_per_row = (int *)malloc(sizeof(int) * spm.nrow);
        memset(new_nnz_per_row, 0, spm.nrow * sizeof(int));
        nzinfo<T> *new_nzentry = (nzinfo<T> *)malloc(sizeof(nzinfo<T>) * ture_nnzero);
        
        int ptr = 0;
        for(int i = 0; i < spm.nnzero; i++)
        {
            if((spm.nzentry)[i].row_num != (spm.nzentry)[i].col_num)
            {
                new_nzentry[ptr].row_num = (spm.nzentry)[i].row_num; 
                new_nnz_per_row[new_nzentry[ptr].row_num]++; 
                new_nzentry[ptr].col_num = (spm.nzentry)[i].col_num;  
                new_nzentry[ptr].val = (spm.nzentry)[i].val;
                ptr++;

                new_nzentry[ptr].row_num = (spm.nzentry)[i].col_num; 
                new_nnz_per_row[new_nzentry[ptr].row_num]++; 
                new_nzentry[ptr].col_num = (spm.nzentry)[i].row_num;
                new_nzentry[ptr].val = (spm.nzentry)[i].val;
                ptr++;
            }else{
                new_nzentry[ptr].row_num = (spm.nzentry)[i].row_num; 
                new_nnz_per_row[new_nzentry[ptr].row_num]++;  
                new_nzentry[ptr].col_num = (spm.nzentry)[i].col_num;  
                new_nzentry[ptr].val = (spm.nzentry)[i].val;
                ptr++;
            }
        }
        free(spm.nnz_per_row);
        free(spm.nzentry);
        spm.nnz_per_row = new_nnz_per_row;
        spm.nzentry = new_nzentry;  
        spm.nnzero = ture_nnzero;
    } 

    return spm;
}

template <class T>
void clear_spmatrix(spmatrix<T> *spm) 
{
    free(spm->nnz_per_row);
    free(spm->nzentry);
}

#endif /* __SPMATRIX_H__ */
