/*
 * CUDA for Sparse Solver
 * Author: Luo Li     li.luo@sub.siat.ac.cn
 */ 

#ifndef __SSOR_PRECON_H__
#define __SSOR_PRECON_H__

#include "blas.h"
#include "cutils.h"

template <class T>
struct ssor_precon
{
    csc_triangle<T> lower;
    T *diag;
    csc_triangle<T> upper;
};

template <class T>
double get_precon_flop(ssor_precon<T> *h_precon)
{
    int n = (h_precon->lower).nrow + 1;
    double flop = 6.0 * n + 2.0 * (h_precon->lower).nnzero + 2.0 * (h_precon->upper).nnzero;
    return flop;
}

template <class T>
void ssor_setup(csc_triangle<T> *csc_lower, spm_diag<T> *spm_dia, csc_triangle<T> *csc_upper, ssor_precon<T> *ssor)
{
    (ssor->lower).nrow = csc_lower->nrow;
    (ssor->lower).ncol = csc_lower->ncol;
    (ssor->lower).nnzero = csc_lower->nnzero;
    (ssor->lower).nrow_per_col = csc_lower->nrow_per_col;
    (ssor->lower).col_ptr = csc_lower->col_ptr;
    (ssor->lower).row_ind = csc_lower->row_ind;
    (ssor->lower).nz_val = csc_lower->nz_val;

    ssor->diag = spm_dia->diag;

    (ssor->upper).nrow = csc_upper->nrow;
    (ssor->upper).ncol = csc_upper->ncol;
    (ssor->upper).nnzero = csc_upper->nnzero;
    (ssor->upper).nrow_per_col = csc_upper->nrow_per_col;
    (ssor->upper).col_ptr = csc_upper->col_ptr;
    (ssor->upper).row_ind = csc_upper->row_ind;
    (ssor->upper).nz_val = csc_upper->nz_val;    
}

template <class T>
void clear_precon(ssor_precon<T> *ssor) 
{
    clear_csc_triangle<T>(&(ssor->lower));
    //free(ssor->diag);
    clear_csc_triangle<T>(&(ssor->upper));
}

template <class T>
ssor_precon<T> d_alloc_precon(ssor_precon<T> *h_precon)
{
    ssor_precon<T> d_precon;  

    d_precon.lower.nrow = (h_precon->lower).nrow;
    d_precon.lower.ncol = (h_precon->lower).ncol;
    d_precon.lower.nnzero = (h_precon->lower).nnzero;
    d_precon.lower.nrow_per_col = (h_precon->lower).nrow_per_col;

    d_precon.upper.nrow = (h_precon->upper).nrow;
    d_precon.upper.ncol = (h_precon->upper).ncol;
    d_precon.upper.nnzero = (h_precon->upper).nnzero;
    d_precon.upper.nrow_per_col = (h_precon->upper).nrow_per_col;

    CUDA_SAFE_CALL(cudaMalloc((void **)&d_precon.lower.col_ptr, (d_precon.lower.ncol + 1) * sizeof(int)));
    CUDA_SAFE_CALL(cudaMalloc((void **)&d_precon.lower.row_ind, d_precon.lower.nnzero * sizeof(int)));
    CUDA_SAFE_CALL(cudaMalloc((void **)&d_precon.lower.nz_val, d_precon.lower.nnzero * sizeof(T)));
    CUDA_SAFE_CALL(cudaMemcpy(d_precon.lower.col_ptr, (h_precon->lower).col_ptr, (d_precon.lower.ncol + 1) * sizeof(int), cudaMemcpyHostToDevice));
    CUDA_SAFE_CALL(cudaMemcpy(d_precon.lower.row_ind, (h_precon->lower).row_ind, d_precon.lower.nnzero * sizeof(int), cudaMemcpyHostToDevice));
    CUDA_SAFE_CALL(cudaMemcpy(d_precon.lower.nz_val, (h_precon->lower).nz_val, d_precon.lower.nnzero * sizeof(T), cudaMemcpyHostToDevice));  

    CUDA_SAFE_CALL(cudaMalloc((void **)&d_precon.upper.col_ptr, (d_precon.upper.ncol + 1) * sizeof(int)));
    CUDA_SAFE_CALL(cudaMalloc((void **)&d_precon.upper.row_ind, d_precon.upper.nnzero * sizeof(int)));
    CUDA_SAFE_CALL(cudaMalloc((void **)&d_precon.upper.nz_val, d_precon.upper.nnzero * sizeof(T)));
    CUDA_SAFE_CALL(cudaMemcpy(d_precon.upper.col_ptr, (h_precon->upper).col_ptr, (d_precon.upper.ncol + 1) * sizeof(int), cudaMemcpyHostToDevice));
    CUDA_SAFE_CALL(cudaMemcpy(d_precon.upper.row_ind, (h_precon->upper).row_ind, d_precon.upper.nnzero * sizeof(int), cudaMemcpyHostToDevice));
    CUDA_SAFE_CALL(cudaMemcpy(d_precon.upper.nz_val, (h_precon->upper).nz_val, d_precon.upper.nnzero * sizeof(T), cudaMemcpyHostToDevice));

    CUDA_SAFE_CALL(cudaMalloc((void **)&d_precon.diag, ((h_precon->lower).nrow + 1) * sizeof(T)));
    CUDA_SAFE_CALL(cudaMemcpy(d_precon.diag, h_precon->diag, ((h_precon->lower).nrow + 1) * sizeof(T), cudaMemcpyHostToDevice));

    return d_precon; 
}

template <class T>
void d_dealloc_precon(ssor_precon<T> *d_precon)
{
    CUDA_SAFE_CALL(cudaFree((d_precon->lower).col_ptr));
    CUDA_SAFE_CALL(cudaFree((d_precon->lower).row_ind));
    CUDA_SAFE_CALL(cudaFree((d_precon->lower).nz_val));
    
    CUDA_SAFE_CALL(cudaFree(d_precon->diag));
    
    CUDA_SAFE_CALL(cudaFree((d_precon->upper).col_ptr));
    CUDA_SAFE_CALL(cudaFree((d_precon->upper).row_ind));
    CUDA_SAFE_CALL(cudaFree((d_precon->upper).nz_val));    
}

template <class T>
void precon_ssor_lower_cpu(ssor_precon<T> *h_precon, T *x, const int n, const T omega)
{
    int *col_ptr = (h_precon->lower).col_ptr;
    int *row_ind = (h_precon->lower).row_ind;
    T *nz_val = (h_precon->lower).nz_val;
    T *diag = h_precon->diag;
    
    int start, dist;
    for(int k = 0; k < n - 1; k++)
    {
        x[k] *= omega / diag[k];
        T x_tmp = x[k];
        start = col_ptr[k];
        dist = col_ptr[k + 1];
        for(int jj = start; jj < dist; jj++)
            x[row_ind[jj]] += x_tmp * nz_val[jj];//(D/w - L) rather than L !!! so use positive
    }
    x[n - 1] *= omega / diag[n - 1];    
}

template <class T>
void precon_ssor_diag_cpu(ssor_precon<T> *h_precon, T *x, const int n, const T omega)
{
    T *diag = h_precon->diag;
    for(int i = 0; i < n; i++)
        x[i] *= diag[i] / omega;
}

template <class T>
void precon_ssor_upper_cpu(ssor_precon<T> *h_precon, T *x, const int n, const T omega)
{
    int *col_ptr = (h_precon->upper).col_ptr;
    int *row_ind = (h_precon->upper).row_ind;
    T *nz_val = (h_precon->upper).nz_val;
    T *diag = h_precon->diag;
    
    int start, dist;
    int ii;
    for(int k = n - 1; k > 0; k--)
    {
        x[k] *= omega / diag[k];
        T x_tmp = x[k];
        ii = n - 1 - k;//col_ptr is forward
        start = col_ptr[ii];
        dist = col_ptr[ii + 1];
        for(int jj = start; jj < dist; jj++)
            x[row_ind[jj]] += x_tmp * nz_val[jj];//(D/w - L) rather than L !!! so use positive
    }
    x[0] *= omega / diag[0];    
}

template <class Matrix_T, class T>
void precon_ssor_cpu(Matrix_T *h_spm, ssor_precon<T> *h_precon, T *r, T *h, const int n)
{
    T omega = OMEGA;
    T *temp = (T *)malloc(sizeof(T) * n);
    std::copy(r, r + n, temp);
    
    precon_ssor_lower_cpu<T>(h_precon, temp, n, omega);
    precon_ssor_diag_cpu<T>(h_precon, temp, n, omega);
    precon_ssor_upper_cpu<T>(h_precon, temp, n, omega);

    std::copy(temp, temp + n, h);
    sca_mul_vec<T>(n, 2.0 - omega, h, h);
    free(temp);
}

template <class Matrix_T, class T>
void precon_cpu(Matrix_T *h_spm, ssor_precon<T> *h_precon, T *r, T *h)
{
    int n = h_spm->ncol;
    #if USE_FLOAT_PRECON
    precon_ssor_cpu<Matrix_T, T>(h_spm, h_precon, r, h, n);
    #else
    precon_ssor_cpu<Matrix_T, T>(h_spm, h_precon, r, h, n);
    #endif
}

template <class Matrix_T, class T>
void precon_ssor_single_gpu(Matrix_T *d_spm, ssor_precon<T> *d_precon, ssor_precon<T> *h_precon, T *r, T *h, const int n)
{
    T omega = OMEGA;
    int dimBlock = DIMBLOCK;
    T *temp = 0;
    CUDA_SAFE_CALL(cudaMalloc((void **)&temp, n * sizeof(T)));

    T *diag = d_precon->diag;

    #if USE_TEXTURE
    bind_vec(diag);
    #endif

    #if USE_CPU_PRECON
    T *h_temp = (T *)malloc(sizeof(T) * n);
    CUDA_SAFE_CALL(cudaMemcpy(h_temp, r, n * sizeof(T), cudaMemcpyDeviceToHost));
    precon_ssor_lower_cpu<T>(h_precon, h_temp, n, omega);
    CUDA_SAFE_CALL(cudaMemcpy(temp, h_temp, n * sizeof(T), cudaMemcpyHostToDevice));
    free(h_temp);
    #else
    int *lower_col_ptr = (d_precon->lower).col_ptr;
    int *lower_row_ind = (d_precon->lower).row_ind;
    T *lower_nz_val = (d_precon->lower).nz_val;    
    cublasXcopy(n, r, 1, temp, 1);
    int n_lower = (d_precon->lower).nrow_per_col;
    const dim3 dimGrid_lower = large_grid(max(n_lower, dimBlock), dimBlock);
    unsigned int dimBlock_lower = min(n_lower, dimBlock);    
    precon_ssor_lower_single_gpu<T><<<dimGrid_lower, dimBlock_lower>>>(lower_col_ptr, lower_row_ind, lower_nz_val, diag, temp, n, omega);
    cudaThreadSynchronize();
    #endif

    const dim3 dimGrid_diag = large_grid(n, dimBlock); 
    unsigned int dimBlock_diag = dimBlock;    
    precon_ssor_diag_single_gpu<T><<<dimGrid_diag, dimBlock_diag>>>(diag, temp, n, omega);
    cudaThreadSynchronize();

    #if USE_CPU_PRECON
    h_temp = (T *)malloc(sizeof(T) * n);
    CUDA_SAFE_CALL(cudaMemcpy(h_temp, temp, n * sizeof(T), cudaMemcpyDeviceToHost));
    precon_ssor_upper_cpu<T>(h_precon, h_temp, n, omega);
    CUDA_SAFE_CALL(cudaMemcpy(h, h_temp, n * sizeof(T), cudaMemcpyHostToDevice));
    free(h_temp);
    #else
    int *upper_col_ptr = (d_precon->upper).col_ptr;
    int *upper_row_ind = (d_precon->upper).row_ind;
    T *upper_nz_val = (d_precon->upper).nz_val;    
    int n_upper = (d_precon->upper).nrow_per_col;
    const dim3 dimGrid_upper = large_grid(max(n_upper, dimBlock), dimBlock);
    unsigned int dimBlock_upper = min(n_upper, dimBlock);    
    precon_ssor_upper_single_gpu<T><<<dimGrid_upper, dimBlock_upper>>>(upper_col_ptr, upper_row_ind, upper_nz_val, diag, temp, n, omega);
    cublasXcopy(n, temp, 1, h, 1);
    #endif

    #if USE_TEXTURE
    unbind_vec(diag);
    #endif   

    cublasXscal(n, 2.0 - omega, h, 1);
    CUDA_SAFE_CALL(cudaFree(temp));
}

template <class Matrix_T, class T>
void precon_single_gpu(Matrix_T *d_spm, ssor_precon<T> *d_precon, ssor_precon<T> *h_precon, T *r, T *h)
{
    int n = d_spm->ncol;
    #if USE_FLOAT_PRECON
    precon_ssor_single_gpu<Matrix_T, T>(d_spm, d_precon, h_precon, r, h, n);
    #else
    precon_ssor_single_gpu<Matrix_T, T>(d_spm, d_precon, h_precon, r, h, n);   
    #endif    
}

#endif /* __SSOR_PRECON_H__ */
