/*
 * CUDA for Sparse Solver
 * Author: Luo Li     li.luo@sub.siat.ac.cn
 */ 

#ifndef __SUB_SSOR_PRECON_H__
#define __SUB_SSOR_PRECON_H__

#include "blas.h"
#include "cutils.h"

template <class T>
struct sub_ssor_precon
{
    csc_tri_array<T> lower;
    T *diag;
    csc_tri_array<T> upper;
};

template <class T>
double get_precon_flop(sub_ssor_precon<T> *h_precon)
{
    int n = (h_precon->lower).nrow;
    int nnzero = (h_precon->lower).nnzero - n;
    double flop = 6.0 * n + 2.0 * nnzero;
    return flop;
}

template <class T>
void ssor_setup(csc_tri_array<T> *csc_lower, spm_diag<T> *spm_dia, csc_tri_array<T> *csc_upper, sub_ssor_precon<T> *ssor)
{
    (ssor->lower).ntri = csc_lower->ntri;
    (ssor->lower).csc_tri_entry = csc_lower->csc_tri_entry;
    
    ssor->diag = spm_dia->diag;

    (ssor->upper).ntri = csc_upper->ntri;
    (ssor->upper).csc_tri_entry = csc_upper->csc_tri_entry;
}

template <class T>
void clear_precon(sub_ssor_precon<T> *ssor) 
{
    clear_csc_triangle<T>(&(ssor->lower));
    free(ssor->diag);
    clear_csc_triangle<T>(&(ssor->upper));
}

template <class T>
sub_ssor_precon<T> d_alloc_precon(sub_ssor_precon<T> *h_precon)
{
    sub_ssor_precon<T> d_precon; 

    d_precon.lower.ntri = (h_precon->lower).ntri;
    d_precon.lower.csc_tri_entry = (csc_triangle<T> *)malloc(sizeof(csc_triangle<T>) * d_precon.lower.ntri);
    for(int i = 0; i < d_precon.lower.ntri; i++)
    {
        csc_triangle<T> *h_csc_tri = &((h_precon->lower).csc_tri_entry[i]);
        csc_triangle<T> *d_csc_tri = &(d_precon.lower.csc_tri_entry[i]);
        d_csc_tri->nrow = h_csc_tri->nrow;
        d_csc_tri->ncol = h_csc_tri->ncol;
        d_csc_tri->nnzero = h_csc_tri->nnzero;
        d_csc_tri->nrow_per_col = h_csc_tri->nrow_per_col;
        CUDA_SAFE_CALL(cudaMalloc((void **)&d_csc_tri->col_ptr, (d_csc_tri->ncol + 1) * sizeof(int)));
        CUDA_SAFE_CALL(cudaMalloc((void **)&d_csc_tri->row_ind, d_csc_tri->nnzero * sizeof(int)));
        CUDA_SAFE_CALL(cudaMalloc((void **)&d_csc_tri->nz_val, d_csc_tri->nnzero * sizeof(T)));
        CUDA_SAFE_CALL(cudaMemcpy(d_csc_tri->col_ptr, h_csc_tri->col_ptr, (d_csc_tri->ncol + 1) * sizeof(int), cudaMemcpyHostToDevice));
        CUDA_SAFE_CALL(cudaMemcpy(d_csc_tri->row_ind, h_csc_tri->row_ind, d_csc_tri->nnzero * sizeof(int), cudaMemcpyHostToDevice));
        CUDA_SAFE_CALL(cudaMemcpy(d_csc_tri->nz_val, h_csc_tri->nz_val, d_csc_tri->nnzero * sizeof(T), cudaMemcpyHostToDevice));
    }
    
    CUDA_SAFE_CALL(cudaMalloc((void **)&d_precon.diag, (h_precon->lower).nrow * sizeof(T)));
    CUDA_SAFE_CALL(cudaMemcpy(d_precon.diag, h_precon->diag, (h_precon->lower).nrow * sizeof(T), cudaMemcpyHostToDevice));
    
    d_precon.upper.ntri = (h_precon->upper).ntri;
    d_precon.upper.csc_tri_entry = (csc_triangle<T> *)malloc(sizeof(csc_triangle<T>) * d_precon.upper.ntri);
    for(int i = 0; i < d_precon.upper.ntri; i++)
    {
        csc_triangle<T> *h_csc_tri = &((h_precon->upper).csc_tri_entry[i]);
        csc_triangle<T> *d_csc_tri = &(d_precon.upper.csc_tri_entry[i]);        
        d_csc_tri->nrow = h_csc_tri->nrow;
        d_csc_tri->ncol = h_csc_tri->ncol;
        d_csc_tri->nnzero = h_csc_tri->nnzero;
        d_csc_tri->nrow_per_col = h_csc_tri->nrow_per_col;
        CUDA_SAFE_CALL(cudaMalloc((void **)&d_csc_tri->col_ptr, (d_csc_tri->ncol + 1) * sizeof(int)));
        CUDA_SAFE_CALL(cudaMalloc((void **)&d_csc_tri->row_ind, d_csc_tri->nnzero * sizeof(int)));
        CUDA_SAFE_CALL(cudaMalloc((void **)&d_csc_tri->nz_val, d_csc_tri->nnzero * sizeof(T)));
        CUDA_SAFE_CALL(cudaMemcpy(d_csc_tri->col_ptr, h_csc_tri->col_ptr, (d_csc_tri->ncol + 1) * sizeof(int), cudaMemcpyHostToDevice));
        CUDA_SAFE_CALL(cudaMemcpy(d_csc_tri->row_ind, h_csc_tri->row_ind, d_csc_tri->nnzero * sizeof(int), cudaMemcpyHostToDevice));
        CUDA_SAFE_CALL(cudaMemcpy(d_csc_tri->nz_val, h_csc_tri->nz_val, d_csc_tri->nnzero * sizeof(T), cudaMemcpyHostToDevice));
    }

    return d_precon; 
}

template <class T>
void d_dealloc_precon(sub_ssor_precon<T> *d_precon)
{
    for(int i = 0; i < (d_precon->lower).ntri; i++)
    {   
        csc_triangle<T> *d_csc_tri = &((d_precon->lower).csc_tri_entry[i]);
        CUDA_SAFE_CALL(cudaFree(d_csc_tri->col_ptr));
        CUDA_SAFE_CALL(cudaFree(d_csc_tri->row_ind));
        CUDA_SAFE_CALL(cudaFree(d_csc_tri->nz_val));
    }
    free((d_precon->lower).csc_tri_entry);

    CUDA_SAFE_CALL(cudaFree(d_precon->diag));
    
    for(int i = 0; i < (d_precon->upper).ntri; i++)
    {    
        csc_triangle<T> *d_csc_tri = &((d_precon->upper).csc_tri_entry[i]);        
        CUDA_SAFE_CALL(cudaFree(d_csc_tri->col_ptr));
        CUDA_SAFE_CALL(cudaFree(d_csc_tri->row_ind));
        CUDA_SAFE_CALL(cudaFree(d_csc_tri->nz_val));  
    }
    free((d_precon->upper).csc_tri_entry);
}

template <class Matrix_T, class T>
void precon_ssor_single_gpu(Matrix_T *d_spm, sub_ssor_precon<T> *d_precon, sub_ssor_precon<T> *h_precon, T *r, T *h, const int n)
{
    T omega = OMEGA;
    int interval_size = INTERVAL_SIZE;	
    unsigned int dimBlock = DIMBLOCK;
    unsigned int rows_per_block = dimBlock / interval_size;	
    
    T *diag = d_precon->diag;
    T *temp1 = 0;
    T *temp2 = 0;   
    CUDA_SAFE_CALL(cudaMalloc((void **)&temp1, n * sizeof(T)));
    CUDA_SAFE_CALL(cudaMalloc((void **)&temp2, n * sizeof(T)));
    cublasXcopy(n, r, 1, temp1, 1);

    for(int i = 0; i < (d_precon->lower).ntri; i++)
    {
        int offset_subtri = i * dimBlock;
        int n_subtri = min(dimBlock, n - offset_subtri);
        int offset_csrmv = offset_subtri + n_subtri;
        int n_csrmv = n - offset_csrmv;

        unsigned int dimBlock_sutri = n_subtri;
        //unsigned int mem_size_subtri = dimBlock_sutri * sizeof(T);

        const dim3 dimGrid = large_grid(max(n_csrmv, rows_per_block), rows_per_block);
        unsigned int dimBlock_csrmv = n_subtri; 
        unsigned int mem_size_csrmv = dimBlock_csrmv * sizeof(T);
        
        csc_triangle<T> *d_csc_tri = &((d_precon->lower).csc_tri_entry[i]);
        #if USE_TEXTURE
        bind_vec(d_csc_tri->nz_val);
        #endif
        ssor_subtri_lower<T><<<1, dimBlock_sutri>>>(d_csc_tri->col_ptr, d_csc_tri->row_ind, d_csc_tri->nz_val, diag, temp1, offset_subtri, n_subtri, n, omega);
        cudaThreadSynchronize();
        #if USE_TEXTURE
        unbind_vec(d_csc_tri->nz_val);
        #endif
        #if USE_TEXTURE
        bind_vec(diag);
        #endif
        ssor_diag<T><<<1, dimBlock_sutri>>>(diag, temp1, temp2, offset_subtri, n_subtri, n, omega);
        ssor_modify_lower<T><<<dimGrid, dimBlock_csrmv, mem_size_csrmv>>>(d_spm->row_ptr, d_spm->col_ind, d_spm->nz_val, temp1, dimBlock_csrmv, offset_csrmv, interval_size, n);
        cudaThreadSynchronize();
        #if USE_TEXTURE
        unbind_vec(diag);
        #endif        
    }
  
    for(int i = 0; i < (d_precon->upper).ntri; i++)
    {
        int offset_subtri = i * dimBlock;
        int n_subtri = min(dimBlock, n - offset_subtri);
        int offset_csrmv = offset_subtri + n_subtri;
        int n_csrmv = n - offset_csrmv;
    
        unsigned int dimBlock_sutri = n_subtri;
        //unsigned int mem_size_subtri = dimBlock_sutri * sizeof(T);
    
        const dim3 dimGrid = large_grid(max(n_csrmv, rows_per_block), rows_per_block);
        unsigned int dimBlock_csrmv = n_subtri; 
        unsigned int mem_size_csrmv = dimBlock_csrmv * sizeof(T);
        
        csc_triangle<T> *d_csc_tri = &((d_precon->upper).csc_tri_entry[i]);
        #if USE_TEXTURE
        bind_vec(d_csc_tri->nz_val);
        #endif
        ssor_subtri_upper<T><<<1, dimBlock_sutri>>>(d_csc_tri->col_ptr, d_csc_tri->row_ind, d_csc_tri->nz_val, diag, temp2, offset_subtri, n_subtri, n, omega);
        cudaThreadSynchronize();
        #if USE_TEXTURE
        unbind_vec(d_csc_tri->nz_val);
        #endif

        ssor_modify_upper<T><<<dimGrid, dimBlock_csrmv, mem_size_csrmv>>>(d_spm->row_ptr, d_spm->col_ind, d_spm->nz_val, temp2, dimBlock_csrmv, offset_csrmv, interval_size, n);
        cudaThreadSynchronize();
    }

    cublasXcopy(n, temp2, 1, h, 1);
    cublasXscal(n, 2.0 - omega, h, 1);
    CUDA_SAFE_CALL(cudaFree(temp1));
    CUDA_SAFE_CALL(cudaFree(temp2));
}

template <class Matrix_T, class T>
void precon_single_gpu(Matrix_T *d_spm, sub_ssor_precon<T> *d_precon, sub_ssor_precon<T> *h_precon, T *r, T *h)
{
    int n = d_spm->ncol;
    #if USE_FLOAT_PRECON
    precon_ssor_single_gpu<Matrix_T, T>(d_spm, d_precon, h_precon, r, h, n);
    #else
    precon_ssor_single_gpu<Matrix_T, T>(d_spm, d_precon, h_precon, r, h, n);   
    #endif    
}

#endif /* __SUB_SSOR_PRECON_H__ */
